"""
functions and utilities for handling zipped protocol bundles
"""
from datetime import date
import json
from pathlib import PurePosixPath, PurePath
from typing import Dict, BinaryIO, TYPE_CHECKING
from zipfile import ZipFile

from opentrons.calibration_storage.helpers import uri_from_definition

from .types import BundleContents

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition

MAIN_PROTOCOL_FILENAME = "protocol.ot2.py"
LABWARE_DIR = "labware"
DATA_DIR = "data"


def _has_files_at_root(zipFile):
    for zipInfo in zipFile.infolist():
        if zipInfo.filename.count("/") == 0:
            return True
    return False


def extract_bundle(bundle: ZipFile) -> BundleContents:  # noqa: C901
    """Extract a bundle and verify its contents and structure."""
    if not _has_files_at_root(bundle):
        raise RuntimeError(
            "No files found in ZIP file's root directory. When selecting "
            "files to zip, make sure to directly select the files "
            "themselves. Do not select their parent directory, which would "
            "result in nesting all files inside that directory in the ZIP."
        )
    try:
        with bundle.open(MAIN_PROTOCOL_FILENAME, "r") as protocol_file:
            py_protocol = protocol_file.read().decode("utf-8")
    except KeyError:
        raise RuntimeError(
            f"Bundled protocol should have a {MAIN_PROTOCOL_FILENAME} "
            + "file in the root directory"
        )
    bundled_labware: Dict[str, "LabwareDefinition"] = {}
    bundled_data = {}
    bundled_python = {}
    for zipInfo in bundle.infolist():
        filepath = PurePosixPath(zipInfo.filename)
        rootpath = filepath.parts[0]

        # skip directories and weird OS-added directories
        # (note: the __MACOSX dir would contain '__MACOSX/foo.py'
        # and other files. This would break our inferences, so we need
        # to exclude all contents of that directory)
        if rootpath == "__MACOSX" or zipInfo.is_dir():
            continue

        with bundle.open(zipInfo) as f:
            if rootpath == LABWARE_DIR and filepath.suffix == ".json":
                labware_def = json.loads(f.read().decode("utf-8"))
                labware_key = uri_from_definition(labware_def)
                if labware_key in bundled_labware:
                    raise RuntimeError(f"Conflicting labware in bundle: {labware_key}")
                bundled_labware[labware_key] = labware_def
            elif rootpath == DATA_DIR:
                # note: data files are read as binary
                bundled_data[str(filepath.relative_to(DATA_DIR))] = f.read()
            elif filepath.suffix == ".py" and str(filepath) != MAIN_PROTOCOL_FILENAME:
                bundled_python[str(filepath)] = f.read().decode("utf-8")

    if not bundled_labware:
        raise RuntimeError("No labware definitions found in bundle.")

    return BundleContents(py_protocol, bundled_labware, bundled_data, bundled_python)


def create_bundle(contents: BundleContents, into_file: BinaryIO):
    """Create a bundle from assumed-good contents"""
    with ZipFile(into_file, mode="w") as zf:
        zf.writestr(MAIN_PROTOCOL_FILENAME, contents.protocol)
        for dataname, datafile in contents.bundled_data.items():
            name = PurePath(dataname).name
            zf.writestr(f"{DATA_DIR}/{name}", datafile)
        for lwdef in contents.bundled_labware.values():
            zipsafe = uri_from_definition(lwdef, "-")
            zf.writestr(f"{LABWARE_DIR}/{zipsafe}.json", json.dumps(lwdef))
        zf.writestr(".bundle_beta", str(date.today()))



"""Run control command side-effect logic."""

from ..state import StateStore, PauseAction


class RunControlHandler:
    """Implementation logic for protocol run control."""

    _state_store: StateStore

    def __init__(self, state_store: StateStore) -> None:
        """Initialize a RunControlHandler instance."""
        self._state_store = state_store

    async def pause(self) -> None:
        """Issue a PauseAction to the store, pausing the run."""
        self._state_store.handle_action(PauseAction())
        await self._state_store.wait_for(
            condition=self._state_store.commands.get_is_running
        )



from __future__ import annotations
import asyncio
import contextlib
import logging
from typing import (
    Dict,
    Iterator,
    List,
    Callable,
    Optional,
    Tuple,
    Union,
    TYPE_CHECKING,
    cast,
)
from collections import OrderedDict

from opentrons.hardware_control import SynchronousAdapter, ThreadManager
from opentrons.hardware_control.modules.types import ModuleType
from opentrons import types
from opentrons.commands import protocol_commands as cmds, types as cmd_types
from opentrons.commands.publisher import CommandPublisher, publish
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.context.protocol_api.instrument_context import (
    InstrumentContextImplementation,
)
from opentrons.protocols.context.simulator.instrument_context import (
    InstrumentContextSimulation,
)
from opentrons.protocols.types import Protocol
from .labware import Labware
from opentrons.protocols.context.labware import AbstractLabware
from opentrons.protocols.context.protocol import AbstractProtocol
from opentrons.protocols.geometry.module_geometry import ModuleGeometry
from opentrons.protocols.geometry.deck import Deck
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from .instrument_context import InstrumentContext
from .module_contexts import (
    ModuleContext,
    MagneticModuleContext,
    TemperatureModuleContext,
    ThermocyclerContext,
)
from opentrons.protocols.api_support.util import (
    AxisMaxSpeeds,
    requires_version,
    APIVersionError,
)

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition

logger = logging.getLogger(__name__)

ModuleTypes = Union[
    "TemperatureModuleContext", "MagneticModuleContext", "ThermocyclerContext"
]


class ProtocolContext(CommandPublisher):
    """The Context class is a container for the state of a protocol.

    It encapsulates many of the methods formerly found in the Robot class,
    including labware, instrument, and module loading, as well as core
    functions like pause and resume.

    Unlike the old robot class, it is designed to be ephemeral. The lifetime
    of a particular instance should be about the same as the lifetime of a
    protocol. The only exception is the one stored in
    :py:attr:`.legacy_api.api.robot`, which is provided only for back
    compatibility and should be used less and less as time goes by.

    .. versionadded:: 2.0

    """

    def __init__(
        self,
        implementation: AbstractProtocol,
        loop: asyncio.AbstractEventLoop = None,
        broker=None,
        api_version: Optional[APIVersion] = None,
    ) -> None:
        """Build a :py:class:`.ProtocolContext`.

        :param loop: An event loop to use. If not specified, this ctor will
                     (eventually) call :py:meth:`asyncio.get_event_loop`.
        :param broker: An optional command broker to link to. If not
                      specified, a dummy one is used.
        :param api_version: The API version to use. If this is ``None``, uses
                            the max supported version.
        """
        super().__init__(broker)

        self._implementation = implementation

        self._api_version = api_version or MAX_SUPPORTED_VERSION
        if self._api_version > MAX_SUPPORTED_VERSION:
            raise RuntimeError(
                f"API version {self._api_version} is not supported by this "
                f"robot software. Please either reduce your requested API "
                f"version or update your robot."
            )
        self._loop = loop or asyncio.get_event_loop()
        self._instruments: Dict[types.Mount, Optional[InstrumentContext]] = {
            mount: None for mount in types.Mount
        }
        self._modules: List[ModuleContext] = []

        self._commands: List[str] = []
        self._unsubscribe_commands: Optional[Callable[[], None]] = None
        self.clear_commands()

    @classmethod
    def build_using(
        cls, implementation: AbstractProtocol, protocol: Protocol, *args, **kwargs
    ) -> ProtocolContext:
        """Build an API instance for the specified parsed protocol

        This is used internally to provision the context with bundle
        contents or api levels.
        """
        kwargs["api_version"] = getattr(protocol, "api_level", MAX_SUPPORTED_VERSION)
        kwargs["implementation"] = implementation
        return cls(*args, **kwargs)

    @property  # type: ignore
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        """Return the API version supported by this protocol context.

        The supported API version was specified when the protocol context
        was initialized. It may be lower than the highest version supported
        by the robot software. For the highest version supported by the
        robot software, see :py:attr:`.protocol_api.MAX_SUPPORTED_VERSION`.
        """
        return self._api_version

    @property
    def _hw_manager(self):
        # TODO (lc 01-05-2021) remove this once we have a more
        # user facing hardware control http api.
        logger.warning(
            "This function will be deprecated in later versions."
            "Please use with caution."
        )
        return self._implementation.get_hardware()

    @property  # type: ignore
    @requires_version(2, 0)
    def bundled_data(self) -> Dict[str, bytes]:
        """Accessor for data files bundled with this protocol, if any.

        This is a dictionary mapping the filenames of bundled datafiles, with
        extensions but without paths (e.g. if a file is stored in the bundle as
        ``data/mydata/aspirations.csv`` it will be in the dict as
        ``'aspirations.csv'``) to the bytes contents of the files.
        """
        return self._implementation.get_bundled_data()

    def cleanup(self):
        """Finalize and clean up the protocol context."""
        if self._unsubscribe_commands:
            self._unsubscribe_commands()
            self._unsubscribe_commands = None

    def __del__(self):
        if getattr(self, "_unsubscribe_commands", None):
            self._unsubscribe_commands()  # type: ignore

    @property  # type: ignore
    @requires_version(2, 0)
    def max_speeds(self) -> AxisMaxSpeeds:
        """Per-axis speed limits when moving this instrument.

        Changing this value changes the speed limit for each non-plunger
        axis of the robot, when moving this pipette. Note that this does
        only sets a limit on how fast movements can be; movements can
        still be slower than this. However, it is useful if you require
        the robot to move much more slowly than normal when using this
        pipette.

        This is a dictionary mapping string names of axes to float values
        limiting speeds. To change a speed, set that axis's value. To
        reset an axis's speed to default, delete the entry for that axis
        or assign it to ``None``.

        For instance,

        .. code-block:: py

            def run(protocol):
                protocol.comment(str(right.max_speeds))  # '{}' - all default
                protocol.max_speeds['A'] = 10  # limit max speed of
                                               # right pipette Z to 10mm/s
                del protocol.max_speeds['A']  # reset to default
                protocol.max_speeds['X'] = 10  # limit max speed of x to
                                               # 10 mm/s
                protocol.max_speeds['X'] = None  # reset to default

        """
        return self._implementation.get_max_speeds()

    @requires_version(2, 0)
    def commands(self):
        return self._commands

    @requires_version(2, 0)
    def clear_commands(self):
        self._commands.clear()
        if self._unsubscribe_commands:
            self._unsubscribe_commands()

        def on_command(message):
            payload = message.get("payload")
            text = payload.get("text")
            if text is None:
                return

            if message["$"] == "before":
                self._commands.append(text.format(**payload))

        self._unsubscribe_commands = self.broker.subscribe(
            cmd_types.COMMAND, on_command
        )

    @contextlib.contextmanager
    def temp_connect(self, hardware: Union[ThreadManager, SynchronousAdapter]):
        """Connect temporarily to the specified hardware controller.

        This should be used as a context manager:

        .. code-block :: python

            with ctx.temp_connect(hw):
                # do some tasks
                ctx.home()
            # after the with block, the context is connected to the same
            # hardware control API it was connected to before, even if
            # an error occurred in the code inside the with block

        """
        hardware_manager = self._implementation.get_hardware()
        old_hw = hardware_manager.hardware
        old_tc = None
        tc_context = None
        # Cache the current instrument context implementations.
        old_instrument_impls = {
            k: v._implementation for k, v in self._instruments.items() if v
        }
        try:
            hardware_manager.set_hw(hardware)
            for mod_ctx in self._modules:
                if isinstance(mod_ctx, ThermocyclerContext):
                    tc_context = mod_ctx
                    hw_tc = next(
                        hw_mod
                        for hw_mod in hardware.attached_modules
                        if hw_mod.name() == "thermocycler"
                    )
                    if hw_tc:
                        old_tc = mod_ctx._module
                        mod_ctx._module = hw_tc

            # InstrumentContextSimulation is an implementation that has no
            # interaction with hardware controller. This is our fast
            # simulation.
            # InstrumentContext objects using an InstrumentContextSimulation
            # must create an InstrumentContextImplementation in order to
            # actually connect the InstrumentContext to the hardware
            # connection.
            for instrument_ctx in self._instruments.values():
                if not instrument_ctx:
                    continue
                impl = instrument_ctx._implementation
                if isinstance(impl, InstrumentContextSimulation):
                    instrument_ctx._implementation = InstrumentContextImplementation(
                        protocol_interface=self._implementation,
                        mount=impl.get_mount(),
                        default_speed=impl.get_default_speed(),
                        instrument_name=impl.get_instrument_name(),
                        api_version=self._api_version,
                    )

            yield self
        finally:
            hardware_manager.set_hw(old_hw)
            if tc_context is not None and old_tc is not None:
                tc_context._module = old_tc
            # reset the instrument context implementations.
            for mount, instrument_impl in old_instrument_impls.items():
                instrument_context = self._instruments[mount]
                if instrument_context:
                    instrument_context._implementation = instrument_impl

    @requires_version(2, 0)
    def connect(self, hardware: Union[ThreadManager, SynchronousAdapter]):
        """Connect to a running hardware API.

        This can be either a simulator or a full hardware controller.

        Note that there is no true disconnected state for a
        :py:class:`.ProtocolContext`; :py:meth:`disconnect` simply creates
        a new simulator and replaces the current hardware with it.
        """
        self._implementation.connect(hardware=hardware)

    @requires_version(2, 0)
    def disconnect(self):
        """Disconnect from currently-connected hardware and simulate instead"""
        self._implementation.disconnect()

    @requires_version(2, 0)
    def is_simulating(self) -> bool:
        return self._implementation.is_simulating()

    @requires_version(2, 0)
    def load_labware_from_definition(
        self,
        labware_def: "LabwareDefinition",
        location: types.DeckLocation,
        label: Optional[str] = None,
    ) -> Labware:
        """Specify the presence of a piece of labware on the OT2 deck.

        This function loads the labware definition specified by `labware_def`
        to the location specified by `location`.

        :param labware_def: The labware definition to load
        :param location: The slot into which to load the labware such as
                         1 or '1'
        :type location: int or str
        :param str label: An optional special name to give the labware. If
                          specified, this is the name the labware will appear
                          as in the run log and the calibration view in the
                          Opentrons app.
        """
        implementation = self._implementation.load_labware_from_definition(
            labware_def=labware_def, location=location, label=label
        )
        return Labware(implementation=implementation)

    @requires_version(2, 0)
    def load_labware(
        self,
        load_name: str,
        location: types.DeckLocation,
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """Load a labware onto the deck given its name.

        For labware already defined by Opentrons, this is a convenient way
        to collapse the two stages of labware initialization (creating
        the labware and adding it to the protocol) into one.

        This function returns the created and initialized labware for use
        later in the protocol.

        :param load_name: A string to use for looking up a labware definition
        :param location: The slot into which to load the labware such as
                         1 or '1'
        :type location: int or str
        :param str label: An optional special name to give the labware. If
                          specified, this is the name the labware will appear
                          as in the run log and the calibration view in the
                          Opentrons app.
        :param str namespace: The namespace the labware definition belongs to.
            If unspecified, will search 'opentrons' then 'custom_beta'
        :param int version: The version of the labware definition. If
            unspecified, will use version 1.
        """
        implementation = self._implementation.load_labware(
            load_name=load_name,
            location=location,
            label=label,
            namespace=namespace,
            version=version,
        )
        return Labware(implementation=implementation)

    @requires_version(2, 0)
    def load_labware_by_name(
        self,
        load_name: str,
        location: types.DeckLocation,
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: int = 1,
    ) -> Labware:
        """
        .. deprecated:: 2.0
            Use :py:meth:`load_labware` instead.
        """
        logger.warning("load_labware_by_name is deprecated. Use load_labware instead.")
        return self.load_labware(load_name, location, label, namespace, version)

    @property  # type: ignore
    @requires_version(2, 0)
    def loaded_labwares(self) -> Dict[int, Labware]:
        """Get the labwares that have been loaded into the protocol context.

        Slots with nothing in them will not be present in the return value.

        .. note::

            If a module is present on the deck but no labware has been loaded
            into it with :py:meth:`.ModuleContext.load_labware`, there will
            be no entry for that slot in this value. That means you should not
            use ``loaded_labwares`` to determine if a slot is available or not,
            only to get a list of labwares. If you want a data structure of all
            objects on the deck regardless of type, see :py:attr:`deck`.


        :returns: Dict mapping deck slot number to labware, sorted in order of
                  the locations.
        """

        def _only_labwares() -> Iterator[Tuple[int, Labware]]:
            for slotnum, slotitem in self._implementation.get_deck().items():
                if isinstance(slotitem, AbstractLabware):
                    yield slotnum, Labware(implementation=slotitem)
                elif isinstance(slotitem, Labware):
                    yield slotnum, slotitem
                elif isinstance(slotitem, ModuleGeometry):
                    if slotitem.labware:
                        yield slotnum, slotitem.labware

        return dict(_only_labwares())

    @requires_version(2, 0)
    def load_module(
        self,
        module_name: str,
        location: Optional[types.DeckLocation] = None,
        configuration: Optional[str] = None,
    ) -> ModuleTypes:
        """Load a module onto the deck given its name.

        This is the function to call to use a module in your protocol, like
        :py:meth:`load_instrument` is the method to call to use an instrument
        in your protocol. It returns the created and initialized module
        context, which will be a different class depending on the kind of
        module loaded.

        A map of deck positions to loaded modules can be accessed later
        using :py:attr:`loaded_modules`.

        :param str module_name: The name or model of the module.
        :param location: The location of the module. This is usually the
                         name or number of the slot on the deck where you
                         will be placing the module. Some modules, like
                         the Thermocycler, are only valid in one deck
                         location. You do not have to specify a location
                         when loading a Thermocycler - it will always be
                         in Slot 7.
        :param configuration: Used to specify the slot configuration of
                              the Thermocycler. Only Valid in Python API
                              Version 2.4 and later. If you wish to use
                              the non-full plate configuration, you must
                              pass in the key word value `semi`
        :type location: str or int or None
        :returns ModuleContext: The loaded and initialized
                                :py:class:`ModuleContext`.
        """
        if self._api_version < APIVersion(2, 4) and configuration:
            raise APIVersionError(
                f"You have specified API {self._api_version}, but you are"
                "using thermocycler parameters only available in 2.4"
            )

        module = self._implementation.load_module(
            module_name=module_name, location=location, configuration=configuration
        )

        if not module:
            raise RuntimeError(f"Could not find specified module: {module_name}")

        mod_class = {
            ModuleType.MAGNETIC: MagneticModuleContext,
            ModuleType.TEMPERATURE: TemperatureModuleContext,
            ModuleType.THERMOCYCLER: ThermocyclerContext,
        }[module.type]

        module_context = mod_class(
            self, module.module, module.geometry, self.api_version, self._loop
        )
        self._modules.append(module_context)
        return module_context

    @property  # type: ignore
    @requires_version(2, 0)
    def loaded_modules(self) -> Dict[int, "ModuleContext"]:
        """Get the modules loaded into the protocol context.

        This is a map of deck positions to modules loaded by previous calls
        to :py:meth:`load_module`. It is not necessarily the same as the
        modules attached to the robot - for instance, if the robot has a
        Magnetic Module and a Temperature Module attached, but the protocol
        has only loaded the Temperature Module with :py:meth:`load_module`,
        only the Temperature Module will be present.

        :returns Dict[str, ModuleContext]: Dict mapping slot name to module
                                           contexts. The elements may not be
                                           ordered by slot number.
        """

        def _modules() -> Iterator[Tuple[int, "ModuleContext"]]:
            for module in self._modules:
                yield int(str(module.geometry.parent)), module

        return OrderedDict(_modules())

    @requires_version(2, 0)
    def load_instrument(
        self,
        instrument_name: str,
        mount: Union[types.Mount, str],
        tip_racks: List[Labware] = None,
        replace: bool = False,
    ) -> "InstrumentContext":
        """Load a specific instrument required by the protocol.

        This value will actually be checked when the protocol runs, to
        ensure that the correct instrument is attached in the specified
        location.

        :param str instrument_name: The name of the instrument model, or a
                                    prefix. For instance, 'p10_single' may be
                                    used to request a P10 single regardless of
                                    the version.
        :param mount: The mount in which this instrument should be attached.
                      This can either be an instance of the enum type
                      :py:class:`.types.Mount` or one of the strings `'left'`
                      and `'right'`.
        :type mount: types.Mount or str
        :param tip_racks: A list of tip racks from which to pick tips if
                          :py:meth:`.InstrumentContext.pick_up_tip` is called
                          without arguments.
        :type tip_racks: List[:py:class:`.Labware`]
        :param bool replace: Indicate that the currently-loaded instrument in
                             `mount` (if such an instrument exists) should be
                             replaced by `instrument_name`.
        """
        if isinstance(mount, str):
            try:
                checked_mount = types.Mount[mount.upper()]
            except KeyError:
                raise ValueError(
                    "If mount is specified as a string, it should be either"
                    "'left' or 'right' (ignoring capitalization, which the"
                    " system strips), not {}".format(mount)
                )
        elif isinstance(mount, types.Mount):
            checked_mount = mount
        else:
            raise TypeError(
                "mount should be either an instance of opentrons.types.Mount"
                " or a string, but is {}.".format(mount)
            )
        logger.info(
            "Trying to load {} on {} mount".format(
                instrument_name, checked_mount.name.lower()
            )
        )

        impl = self._implementation.load_instrument(
            instrument_name=instrument_name, mount=checked_mount, replace=replace
        )

        new_instr = InstrumentContext(
            ctx=self,
            broker=self.broker,
            implementation=impl,
            at_version=self.api_version,
            tip_racks=tip_racks,
        )
        self._instruments[checked_mount] = new_instr
        return new_instr

    @property  # type: ignore
    @requires_version(2, 0)
    def loaded_instruments(self) -> Dict[str, "InstrumentContext"]:
        """Get the instruments that have been loaded into the protocol.

        This is a map of mount name to instruments previously loaded with
        :py:meth:`load_instrument`. It is not necessarily the same as the
        instruments attached to the robot - for instance, if the robot has
        an instrument in both mounts but your protocol has only loaded one
        of them with :py:meth:`load_instrument`, the unused one will not
        be present.

        :returns: A dict mapping mount name
                  (``'left'`` or ``'right'``)
                  to the instrument in that mount.
                  If a mount has no loaded instrument,
                  that key will be missing from the dict.
        """
        return {
            mount.name.lower(): instr
            for mount, instr in self._instruments.items()
            if instr
        }

    @publish(command=cmds.pause)
    @requires_version(2, 0)
    def pause(self, msg=None) -> None:
        """Pause execution of the protocol until it's resumed.

        A human can resume the protocol through the Opentrons App.

        This function returns immediately, but the next function call that
        is blocked by a paused robot (anything that involves moving) will
        not return until the protocol is resumed.

        :param str msg: An optional message to show to connected clients. The
            Opentrons App will show this in the run log.
        """
        self._implementation.pause(msg=msg)

    @publish(command=cmds.resume)
    @requires_version(2, 0)
    def resume(self) -> None:
        """Resume the protocol after :py:meth:`pause`.

        .. deprecated:: 2.12
           The Python Protocol API supports no safe way for a protocol to resume itself.
           See https://github.com/Opentrons/opentrons/issues/8209.
           If you're looking for a way for your protocol to resume automatically
           after a period of time, use :py:meth:`delay`.
        """
        self._implementation.resume()

    @publish(command=cmds.comment)
    @requires_version(2, 0)
    def comment(self, msg) -> None:
        """
        Add a user-readable comment string that will be echoed to the Opentrons
        app.

        The value of the message is computed during protocol simulation,
        so cannot be used to communicate real-time information from the robot's
        actual run.
        """
        self._implementation.comment(msg=msg)

    @publish(command=cmds.delay)
    @requires_version(2, 0)
    def delay(self, seconds=0, minutes=0, msg=None):
        """Delay protocol execution for a specific amount of time.

        :param float seconds: A time to delay in seconds
        :param float minutes: A time to delay in minutes

        If both `seconds` and `minutes` are specified, they will be added.
        """
        delay_time = seconds + minutes * 60
        self._implementation.delay(seconds=delay_time, msg=msg)

    @requires_version(2, 0)
    def home(self):
        """Homes the robot."""
        logger.debug("home")
        self._implementation.home()

    @property
    def location_cache(self) -> Optional[types.Location]:
        """The cache used by the robot to determine where it last was."""
        return self._implementation.get_last_location()

    @location_cache.setter
    def location_cache(self, loc: Optional[types.Location]):
        self._implementation.set_last_location(loc)

    @property  # type: ignore
    @requires_version(2, 0)
    def deck(self) -> Deck:
        """The object holding the deck layout of the robot.

        This object behaves like a dictionary with keys for both numeric
        and string slot numbers (for instance, ``protocol.deck[1]`` and
        ``protocol.deck['1']`` will both return the object in slot 1). If
        nothing is loaded into a slot, ``None`` will be present. This object
        is useful for determining if a slot in the deck is free. Rather than
        filtering the objects in the deck map yourself, you can also use
        :py:attr:`loaded_labwares` to see a dict of labwares and
        :py:attr:`loaded_modules` to see a dict of modules. For advanced
        control you can delete an item of labware from the deck with
        e.g. ``del protocol.deck['1']`` to free a slot for new labware.
        (Note that for each slot only the last labware used in a command will
        be available for calibration in the OpenTrons UI, and that the
        tallest labware on the deck will be calculated using only currently
        loaded labware, meaning that the labware loaded should always
        reflect the labware physically on the deck (or be higher than the
        labware on the deck).
        """
        return self._implementation.get_deck()

    @property  # type: ignore
    @requires_version(2, 0)
    def fixed_trash(self) -> Labware:
        """The trash fixed to slot 12 of the robot deck.

        It has one well and should be accessed like labware in your protocol.
        e.g. ``protocol.fixed_trash['A1']``
        """
        trash = self._implementation.get_fixed_trash()
        # TODO AL 20201113 - remove this when DeckLayout only holds
        #  LabwareInterface instances.
        if isinstance(trash, AbstractLabware):
            return Labware(implementation=trash)
        return cast("Labware", trash)

    @requires_version(2, 5)
    def set_rail_lights(self, on: bool):
        """
        Controls the robot rail lights

        :param bool on: If true, turn on rail lights; otherwise, turn off.
        """
        self._implementation.set_rail_lights(on=on)

    @property  # type: ignore
    @requires_version(2, 5)
    def rail_lights_on(self) -> bool:
        """Returns True if the rail lights are on"""
        return self._implementation.get_rail_lights_on()

    @property  # type: ignore
    @requires_version(2, 5)
    def door_closed(self) -> bool:
        """Returns True if the robot door is closed"""
        return self._implementation.door_closed()



"""
This was initially generated by datamodel-codegen from the labware schema in
shared-data. It's been modified by hand to be more friendly.

TODO: 20210330 Amit - consider moving this to opentrons-shared-data.
"""

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, Extra, Field
from typing_extensions import Literal

SAFE_STRING_REGEX = "^[a-z0-9._]+$"


class CornerOffsetFromSlot(BaseModel):
    """
    Distance from left-front-bottom corner of slot to left-front-bottom corner
     of labware bounding box. Used for labware that spans multiple slots. For
      labware that does not span multiple slots, x/y/z should all be zero.
    """

    x: float
    y: float
    z: float


class BrandData(BaseModel):
    brand: str = Field(..., description="Brand/manufacturer name")
    brandId: Optional[List[str]] = Field(
        None,
        description="An array of manufacture numbers pertaining to a given labware",
    )
    links: Optional[List[str]] = Field(
        None, description="URLs for manufacturer page(s)"
    )


class DisplayCategory(str, Enum):
    tipRack = "tipRack"
    tubeRack = "tubeRack"
    reservoir = "reservoir"
    trash = "trash"
    wellPlate = "wellPlate"
    aluminumBlock = "aluminumBlock"
    other = "other"


class Metadata(BaseModel):
    """
    Properties used for search and display
    """

    displayName: str = Field(..., description="Easy to remember name of labware")
    displayCategory: DisplayCategory = Field(
        ..., description="Label(s) used in UI to categorize labware"
    )
    displayVolumeUnits: Literal["µL", "mL", "L"] = Field(
        ..., description="Volume units for display"
    )
    tags: Optional[List[str]] = Field(
        None, description="List of descriptions for a given labware"
    )


class Parameters(BaseModel):
    """
    Internal describers used to determine pipette movement to labware
    """

    format: Literal[
        "96Standard", "384Standard", "trough", "irregular", "trash"
    ] = Field(
        ..., description="Property to determine compatibility with multichannel pipette"
    )
    quirks: Optional[List[str]] = Field(
        None,
        description="Property to classify a specific behavior this labware "
        "should have",
    )
    isTiprack: bool = Field(
        ..., description="Flag marking whether a labware is a tiprack or not"
    )
    tipLength: Optional[float] = Field(
        None,
        ge=0.0,
        description="Required if labware is tiprack, specifies length of tip"
        " from drawing or as measured with calipers",
    )
    tipOverlap: Optional[float] = Field(
        None,
        ge=0.0,
        description="Required if labware is tiprack, specifies the length of "
        "the area of the tip that overlaps the nozzle of the pipette",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition",
        regex=SAFE_STRING_REGEX,
    )
    isMagneticModuleCompatible: bool = Field(
        ...,
        description="Flag marking whether a labware is compatible by default "
        "with the Magnetic Module",
    )
    magneticModuleEngageHeight: Optional[float] = Field(
        None, ge=0.0, description="Distance to move magnetic module magnets to engage"
    )


class Dimensions(BaseModel):
    """
    Outer dimensions of a labware
    """

    yDimension: float = Field(..., ge=0.0)
    zDimension: float = Field(..., ge=0.0)
    xDimension: float = Field(..., ge=0.0)


class WellDefinition(BaseModel):
    class Config:
        extra = Extra.allow

    depth: float = Field(..., ge=0.0)
    x: float = Field(
        ...,
        ge=0.0,
        description="x location of center-bottom of well in reference to "
        "left-front-bottom of labware",
    )
    y: float = Field(
        ...,
        ge=0.0,
        description="y location of center-bottom of well in reference to "
        "left-front-bottom of labware",
    )
    z: float = Field(
        ...,
        ge=0.0,
        description="z location of center-bottom of well in reference to "
        "left-front-bottom of labware",
    )
    totalLiquidVolume: float = Field(
        ..., ge=0.0, description="Total well, tube, or tip volume in microliters"
    )
    xDimension: Optional[float] = Field(
        None, ge=0.0, description="x dimension of rectangular wells"
    )
    yDimension: Optional[float] = Field(
        None, ge=0.0, description="y dimension of rectangular wells"
    )
    diameter: Optional[float] = Field(
        None, ge=0.0, description="diameter of circular wells"
    )
    shape: Literal["rectangular", "circular"] = Field(
        ...,
        description="If 'rectangular', use xDimension and "
        "yDimension; if 'circular' use diameter",
    )


class Metadata1(BaseModel):
    """
    Metadata specific to a grid of wells in a labware
    """

    displayName: Optional[str] = Field(
        None, description="User-readable name for the well group"
    )
    displayCategory: Optional[DisplayCategory] = Field(
        None, description="Label(s) used in UI to categorize well groups"
    )
    wellBottomShape: Optional[Literal["flat", "u", "v"]] = Field(
        None, description="Bottom shape of the well for UI purposes"
    )


class Group(BaseModel):
    wells: List[str] = Field(
        ..., description="An array of wells that contain the same metadata", min_items=1
    )
    metadata: Metadata1 = Field(
        ..., description="Metadata specific to a grid of wells in a labware"
    )
    brand: Optional[BrandData] = Field(
        None, description="Brand data for the well group (e.g. for tubes)"
    )


class LabwareDefinition(BaseModel):
    schemaVersion: Literal[1, 2] = Field(
        ..., description="Which schema version a labware is using"
    )
    version: int = Field(
        ...,
        description="Version of the labware definition itself "
        "(eg myPlate v1/v2/v3). An incrementing integer",
        ge=1.0,
    )
    namespace: str = Field(..., regex=SAFE_STRING_REGEX)
    metadata: Metadata = Field(
        ..., description="Properties used for search and display"
    )
    brand: BrandData = Field(
        ...,
        description="Real-world labware that the definition is modeled "
        "from and/or compatible with",
    )
    parameters: Parameters = Field(
        ...,
        description="Internal describers used to determine pipette movement "
        "to labware",
    )
    ordering: List[List[str]] = Field(
        ...,
        description="Generated array that keeps track of how wells should be "
        "ordered in a labware",
    )
    cornerOffsetFromSlot: CornerOffsetFromSlot = Field(
        ...,
        description="Distance from left-front-bottom corner of slot to "
        "left-front-bottom corner of labware bounding box. Used for "
        "labware that spans multiple slots. For labware that does "
        "not span multiple slots, x/y/z should all be zero.",
    )
    dimensions: Dimensions = Field(..., description="Outer dimensions of a labware")
    wells: Dict[str, WellDefinition] = Field(
        ...,
        description="Unordered object of well objects with position and "
        "dimensional information",
    )
    groups: List[Group] = Field(
        ...,
        description="Logical well groupings for metadata/display purposes; "
        "changes in groups do not affect protocol execution",
    )



"""Dispense command request, result, and implementation models."""
from __future__ import annotations
from typing import Optional, Type
from typing_extensions import Literal

from .pipetting_common import BaseLiquidHandlingData, BaseLiquidHandlingResult
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest


DispenseCommandType = Literal["dispense"]


class DispenseData(BaseLiquidHandlingData):
    """Data required to dispense to a specific well."""

    pass


class DispenseResult(BaseLiquidHandlingResult):
    """Result data from the execution of a DispenseRequest."""

    pass


class DispenseImplementation(AbstractCommandImpl[DispenseData, DispenseResult]):
    """Dispense command implementation."""

    async def execute(self, data: DispenseData) -> DispenseResult:
        """Move to and dispense to the requested well."""
        volume = await self._pipetting.dispense(
            pipette_id=data.pipetteId,
            labware_id=data.labwareId,
            well_name=data.wellName,
            well_location=data.wellLocation,
            volume=data.volume,
        )

        return DispenseResult(volume=volume)


class Dispense(BaseCommand[DispenseData, DispenseResult]):
    """Dispense command model."""

    commandType: DispenseCommandType = "dispense"
    data: DispenseData
    result: Optional[DispenseResult]

    _ImplementationCls: Type[DispenseImplementation] = DispenseImplementation


class DispenseRequest(BaseCommandRequest[DispenseData]):
    """Create dispense command request model."""

    commandType: DispenseCommandType = "dispense"
    data: DispenseData

    _CommandCls: Type[Dispense] = Dispense



"""Load labware command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type
from typing_extensions import Literal

from opentrons.protocols.models import LabwareDefinition

from ..types import LabwareLocation, CalibrationOffset
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest

LoadLabwareCommandType = Literal["loadLabware"]


class LoadLabwareData(BaseModel):
    """Data required to load a labware into a slot."""

    location: LabwareLocation = Field(
        ...,
        description="Location the labware should be loaded into.",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The labware definition version.",
    )
    labwareId: Optional[str] = Field(
        None,
        description="An optional ID to assign to this labware. If None, an ID "
        "will be generated.",
    )


class LoadLabwareResult(BaseModel):
    """Result data from the execution of a LoadLabwareRequest."""

    labwareId: str = Field(
        ...,
        description="An ID to reference this labware in subsequent commands.",
    )
    definition: LabwareDefinition = Field(
        ...,
        description="The full definition data for this labware.",
    )
    calibration: CalibrationOffset = Field(
        ...,
        description="Calibration offset data for this labware at load time.",
    )


class LoadLabwareImplementation(
    AbstractCommandImpl[LoadLabwareData, LoadLabwareResult]
):
    """Load labware command implementation."""

    async def execute(self, data: LoadLabwareData) -> LoadLabwareResult:
        """Load definition and calibration data necessary for a labware."""
        loaded_labware = await self._equipment.load_labware(
            load_name=data.loadName,
            namespace=data.namespace,
            version=data.version,
            location=data.location,
            labware_id=data.labwareId,
        )
        x_offset, y_offset, z_offset = loaded_labware.calibration

        return LoadLabwareResult(
            labwareId=loaded_labware.labware_id,
            definition=loaded_labware.definition,
            calibration=CalibrationOffset(x=x_offset, y=y_offset, z=z_offset),
        )


class LoadLabware(BaseCommand[LoadLabwareData, LoadLabwareResult]):
    """Load labware command resource model."""

    commandType: LoadLabwareCommandType = "loadLabware"
    data: LoadLabwareData
    result: Optional[LoadLabwareResult]

    _ImplementationCls: Type[LoadLabwareImplementation] = LoadLabwareImplementation


class LoadLabwareRequest(BaseCommandRequest[LoadLabwareData]):
    """Load labware command creation request."""

    commandType: LoadLabwareCommandType = "loadLabware"
    data: LoadLabwareData

    _CommandCls: Type[LoadLabware] = LoadLabware



from typing import Any, List


def session_listify(location: Any) -> List:
    if isinstance(location, list):
        try:
            return session_listify(location[0])
        except IndexError:
            return [location]
    else:
        return [location]


def from_list(commands):
    """
    Given a list of tuples of form (depth, text)
    that represents a DFS traversal of a command tree,
    returns a dictionary representing command tree.
    """

    def subtrees(commands, level):
        if not commands:
            return

        acc = []
        parent, *commands = commands

        for command in commands:
            if command["level"] > level:
                acc.append(command)
            else:
                yield (parent, acc)
                parent = command
                acc.clear()
        yield (parent, acc)

    def walk(commands, level=0):
        return [
            {
                "description": key["description"],
                "children": walk(subtree, level + 1),
                "id": key["id"],
            }
            for key, subtree in subtrees(commands, level)
        ]

    return walk(commands)



from pydantic import BaseSettings, BaseModel


class PipetteSettings(BaseModel):
    model: str = "p20_single_v2.0"
    id: str = "P20SV202020070101"


class SmoothieSettings(BaseModel):
    left: PipetteSettings = PipetteSettings(
        model="p20_multi_v2.0", id="P3HMV202020041605"
    )
    right: PipetteSettings = PipetteSettings(
        model="p20_single_v2.0", id="P20SV202020070101"
    )


class Settings(BaseSettings):
    smoothie: SmoothieSettings = SmoothieSettings()

    host: str = "0.0.0.0"

    class Config:
        env_prefix = "OT_EMULATOR_"



"""Base command data model and type definitions."""
from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum
from datetime import datetime
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

# convenience type alias to work around type-only circular dependency
if TYPE_CHECKING:
    from opentrons.protocol_engine import execution

CommandDataT = TypeVar("CommandDataT", bound=BaseModel)

CommandResultT = TypeVar("CommandResultT", bound=BaseModel)


class CommandStatus(str, Enum):
    """Command execution status."""

    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class BaseCommandRequest(GenericModel, Generic[CommandDataT]):
    """Base class for command creation requests.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    data: CommandDataT = Field(..., description="Command execution data payload")


class BaseCommand(GenericModel, Generic[CommandDataT, CommandResultT]):
    """Base command model.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    id: str = Field(..., description="Unique identifier for a particular command")
    createdAt: datetime = Field(..., description="Command creation timestamp")
    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    status: CommandStatus = Field(..., description="Command execution status")
    data: CommandDataT = Field(..., description="Command execution data payload")
    result: Optional[CommandResultT] = Field(
        None,
        description="Command execution result data, if succeeded",
    )
    # TODO(mc, 2021-06-08): model ProtocolEngine errors
    error: Optional[str] = Field(
        None,
        description="Command execution failure, if failed",
    )
    startedAt: Optional[datetime] = Field(
        None,
        description="Command execution start timestamp, if started",
    )
    completedAt: Optional[datetime] = Field(
        None,
        description="Command execution completed timestamp, if completed",
    )


class AbstractCommandImpl(
    ABC,
    Generic[CommandDataT, CommandResultT],
):
    """Abstract command creation and execution implementation.

    A given command request should map to a specific command implementation,
    which defines how to:

    - Create a command resource from the request model
    - Execute the command, mapping data from execution into the result model
    """

    def __init__(
        self,
        equipment: execution.EquipmentHandler,
        movement: execution.MovementHandler,
        pipetting: execution.PipettingHandler,
        run_control: execution.RunControlHandler,
    ) -> None:
        """Initialize the command implementation with execution handlers."""
        self._equipment = equipment
        self._movement = movement
        self._pipetting = pipetting
        self._run_control = run_control

    @abstractmethod
    async def execute(self, data: CommandDataT) -> CommandResultT:
        """Execute the command, mapping data from execution into a response model."""
        ...



"""Interfaces to provide data and other external system resources.

Classes in this module do not maintain state and can be instantiated
as needed. Some classes may contain solely static methods.
"""
from .model_utils import ModelUtils
from .deck_data_provider import DeckDataProvider, DeckFixedLabware
from .labware_data_provider import LabwareDataProvider


__all__ = [
    "ModelUtils",
    "LabwareDataProvider",
    "DeckDataProvider",
    "DeckFixedLabware",
]



import typing

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from os import PathLike

from typing_extensions import Literal
from opentrons_shared_data.pipette.dev_types import LabwareUri

CalibrationID = typing.NewType("CalibrationID", str)
StrPath = typing.Union[str, PathLike]
AttitudeMatrix = typing.List[typing.List[float]]
PipetteOffset = typing.List[float]


class SourceType(str, Enum):
    """Calibration source type"""

    default = "default"
    factory = "factory"
    user = "user"
    calibration_check = "calibration_check"
    legacy = "legacy"
    unknown = "unknown"


class TipLengthCalNotFound(Exception):
    pass


@dataclass
class CalibrationStatus:
    markedBad: bool = False
    source: typing.Optional[SourceType] = None
    markedAt: typing.Optional[datetime] = None


@dataclass
class UriDetails:
    namespace: str
    load_name: str
    version: int


@dataclass
class OffsetData:
    """
    Class to categorize the shape of a
    given calibration data.
    """

    value: typing.List[float]
    last_modified: typing.Optional[datetime]


@dataclass
class TipLengthData:
    """
    Class to categorize the shape of a
    given calibration data.
    """

    value: typing.Optional[float] = None
    last_modified: typing.Optional[datetime] = None


@dataclass
class ParentOptions:
    """
    Class to store whether a labware calibration has
    a module, as well the original parent (slot).
    As of now, the slot is not saved in association
    with labware calibrations.

    The slot value will be the empty string.
    """

    slot: str
    module: str = ""


@dataclass
class LabwareCalibrationTypes:
    """
    Class to categorize what calibration
    data might be stored for a labware.
    """

    offset: OffsetData
    tip_length: TipLengthData


@dataclass
class CalibrationInformation:
    """
    Class to store important calibration
    info for labware.
    """

    calibration: LabwareCalibrationTypes
    parent: ParentOptions
    labware_id: str
    uri: str


@dataclass
class TipLengthCalibration:
    tip_length: float
    source: SourceType
    status: CalibrationStatus
    pipette: str
    tiprack: str
    last_modified: datetime
    uri: typing.Union[LabwareUri, Literal[""]]


@dataclass
class DeckCalibration:
    attitude: AttitudeMatrix
    source: SourceType
    status: CalibrationStatus
    last_modified: typing.Optional[datetime] = None
    pipette_calibrated_with: typing.Optional[str] = None
    tiprack: typing.Optional[str] = None


@dataclass
class PipetteOffsetByPipetteMount:
    """
    Class to store pipette offset without pipette and mount info
    """

    offset: PipetteOffset
    source: SourceType
    status: CalibrationStatus
    tiprack: typing.Optional[str] = None
    uri: typing.Optional[str] = None
    last_modified: typing.Optional[datetime] = None


@dataclass
class PipetteOffsetCalibration:
    """
    Class to store pipette offset calibration with pipette and mount info
    """

    pipette: str
    mount: str
    offset: PipetteOffset
    tiprack: str
    uri: str
    last_modified: datetime
    source: SourceType
    status: CalibrationStatus



""" opentrons.calibration_storage.delete: functions that
remove single or multiple calibration files from the
file system.
"""
from pathlib import Path

from . import types as local_types, file_operators as io

from opentrons import config
from opentrons.types import Mount


def clear_calibrations():
    """
    Delete all calibration files for labware. This includes deleting tip-length
    data for tipracks.
    """
    calibration_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    try:
        targets = [f for f in calibration_path.iterdir() if f.suffix == ".json"]
        for target in targets:
            target.unlink()
    except FileNotFoundError:
        pass


def _remove_offset_from_index(calibration_id: local_types.CalibrationID):
    """
    Helper function to remove an individual offset file.

    :param calibration_id: labware hash
    :raises FileNotFoundError: If index file does not exist or
    the specified id is not in the index file.
    """
    offset_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    index_path = offset_path / "index.json"
    blob = io.read_cal_file(str(index_path))

    del blob["data"][calibration_id]
    io.save_to_file(index_path, blob)


def delete_offset_file(calibration_id: local_types.CalibrationID):
    """
    Given a labware's hash, delete the file and remove it from the index file.

    :param calibration_id: labware hash
    """
    offset_path = config.get_opentrons_path("labware_calibration_offsets_dir_v2")
    offset = offset_path / f"{calibration_id}.json"
    try:
        _remove_offset_from_index(calibration_id)
        offset.unlink()
    except FileNotFoundError:
        pass


def _remove_tip_length_from_index(tiprack: str, pipette: str):
    """
    Remove tip length data from the index file
    """
    tip_length_dir = config.get_tip_length_cal_path()
    index_path = tip_length_dir / "index.json"
    blob = io.read_cal_file(str(index_path))

    if tiprack in blob and pipette in blob[tiprack]:
        blob[tiprack].remove(pipette)
        io.save_to_file(index_path, blob)


def delete_tip_length_calibration(tiprack: str, pipette: str):
    """
    Delete tip length calibration based on tiprack hash and
    pipette serial number

    :param tiprack: tiprack hash
    :param pipette: pipette serial number
    """
    tip_length_dir = config.get_tip_length_cal_path()
    tip_length_path = tip_length_dir / f"{pipette}.json"
    blob = io.read_cal_file(str(tip_length_path))

    if tiprack in blob:
        del blob[tiprack]
        if blob:
            io.save_to_file(tip_length_path, blob)
        else:
            tip_length_path.unlink()
        _remove_tip_length_from_index(tiprack, pipette)


def clear_tip_length_calibration():
    """
    Delete all tip length calibration files.
    """
    tip_length_path = config.get_tip_length_cal_path()
    try:
        targets = (f for f in tip_length_path.iterdir() if f.suffix == ".json")
        for target in targets:
            target.unlink()
    except FileNotFoundError:
        pass


def _remove_pipette_offset_from_index(pipette: str, mount: Mount):
    """
    Helper function to remove an individual pipette offset file.

    :param pipette: pipette serial number
    :param mount: pipette mount
    :raises FileNotFoundError: If index file does not exist or
    the specified id is not in the index file.
    """
    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    index_path = offset_dir / "index.json"
    blob = io.read_cal_file(str(index_path))

    try:
        blob[mount.name.lower()].remove(pipette)
        io.save_to_file(index_path, blob)
    except (KeyError, ValueError):
        # If the index file does not have a mount entry, you get
        # an error here
        pass


def delete_pipette_offset_file(pipette: str, mount: Mount):
    """
    Delete pipette offset file based on mount and pipette serial number

    :param pipette: pipette serial number
    :param mount: pipette mount
    """
    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    offset_path = offset_dir / mount.name.lower() / f"{pipette}.json"

    try:
        _remove_pipette_offset_from_index(pipette, mount)
        offset_path.unlink()
    except FileNotFoundError:
        pass


def clear_pipette_offset_calibrations():
    """
    Delete all pipette offset calibration files.
    """

    def _remove_json_files_in_directories(p: Path):
        for item in p.iterdir():
            if item.is_dir():
                _remove_json_files_in_directories(item)
            elif item.suffix == ".json":
                item.unlink()

    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    try:
        _remove_json_files_in_directories(offset_dir)
    except FileNotFoundError:
        pass


def delete_robot_deck_attitude():
    """
    Delete the robot deck attitude calibration.
    """

    robot_dir = config.get_opentrons_path("robot_calibration_dir")
    gantry_path = robot_dir / "deck_calibration.json"
    if gantry_path.exists():
        gantry_path.unlink()



"""Unique ID generation provider."""
from datetime import datetime, timezone
from uuid import uuid4


class ModelUtils:
    """Common resource model utilities provider."""

    @staticmethod
    def generate_id() -> str:
        """Generate a unique identifier.

        Uses UUIDv4 for safety in a multiprocessing environment.
        """
        return str(uuid4())

    @staticmethod
    def get_timestamp() -> datetime:
        """Get a timestamp of the current time."""
        return datetime.now(tz=timezone.utc)



from datetime import timedelta

from . import types as command_types


def comment(msg: str) -> command_types.CommentCommand:
    text = msg
    return {"name": command_types.COMMENT, "payload": {"text": text}}


def delay(
    seconds: float, minutes: float, msg: str = None
) -> command_types.DelayCommand:
    td = timedelta(minutes=minutes, seconds=seconds)
    actual_min, actual_sec = divmod(td.total_seconds(), 60)

    text = (
        f"Delaying for {int(actual_min)} minutes and " f"{round(actual_sec, 3)} seconds"
    )

    if msg:
        text = f"{text}. {msg}"

    return {
        "name": command_types.DELAY,
        "payload": {"minutes": actual_min, "seconds": actual_sec, "text": text},
    }


def pause(msg: str = None) -> command_types.PauseCommand:
    text = "Pausing robot operation"
    if msg:
        text = text + ": {}".format(msg)
    return {
        "name": command_types.PAUSE,
        "payload": {
            "text": text,
            "userMessage": msg,
        },
    }


def resume() -> command_types.ResumeCommand:
    return {
        "name": command_types.RESUME,
        "payload": {"text": "Resuming robot operation"},
    }



import logging
from typing import Optional, List
from typing_extensions import Final

import math
import functools

from dataclasses import dataclass

from opentrons.protocols.geometry.deck import Deck
from opentrons.commands import types
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.duration.errors import DurationEstimatorException
from opentrons.types import Location


# We refer to page 3 of the GEN2 Temperature Module White-Paper
# https://blog.opentrons.com/opentrons-technical-documentation/
# Through the data we notice that there are different
# rates of Celsius/second depending on temperature range.
# These were all tested to be ~95% consistent with the data
TEMP_MOD_RATE_HIGH_AND_ABOVE: Final = 0.3611111111
TEMP_MOD_RATE_LOW_TO_HIGH: Final = 0.2
TEMP_MOD_RATE_ZERO_TO_LOW: Final = 0.0875
TEMP_MOD_LOW_THRESH: Final = 25.0
TEMP_MOD_HIGH_THRESH: Final = 37.0

THERMO_LOW_THRESH: Final = 23.0
THERMO_HIGH_THRESH: Final = 70.0

START_MODULE_TEMPERATURE: Final = 25.0

STARTING_SLOT: Final[str] = "12"


logger = logging.getLogger(__name__)


@dataclass
class TimerEntry:
    command: types.CommandMessage
    duration: float


# TODO (al, 2021-09-09): Write doc strings for each method detailing how the
#  duration is computed.


class DurationEstimator:
    """
    Broker listener that calculates the duration of protocol steps.
    """

    def __init__(self):
        # Which slot the last command was in.
        self._last_deckslot = STARTING_SLOT
        # todo(mm, 2021-09-08): Support protocols with more than one
        # Temperature or Thermocycler Module.
        self._last_temperature_module_temperature = START_MODULE_TEMPERATURE
        self._last_thermocycler_module_temperature = START_MODULE_TEMPERATURE
        # Per step time estimate.
        self._increments: List[TimerEntry] = []
        self._deck = Deck()

    def get_total_duration(self) -> float:
        """Return the total duration"""
        return functools.reduce(
            lambda acc, val: acc + val.duration, self._increments, 0.0
        )

    def on_message(self, message: types.CommandMessage) -> None:
        """
        Message handler for Broker events.

        Args:
            message: A protocol message

        Returns:
            None
        """
        # Whether this message comes before or after the command is being executed..
        if message["$"] != "after":
            # We only want to process the afters.
            return

        # Extract the message name
        message_name = message["name"]
        # The actual payload of the command that varies by message.
        payload = message["payload"]

        # We cannot handle paired pipette messages
        # locations also applies to transfer commands
        if "instruments" in payload:
            logger.warning(
                f"Paired pipettes are not supported by the duration estimator. "
                f"Command '{payload['text']}' cannot be estimated properly."
            )
            return

        location = payload.get("location")

        try:
            duration = self.handle_message(message_name, payload)
        except Exception as e:
            raise DurationEstimatorException(str(e))

        if location:
            slot = self.get_slot(location)
            self._last_deckslot = slot if slot else self._last_deckslot

        self._increments.append(
            TimerEntry(
                command=message,
                duration=duration,
            )
        )

    def handle_message(  # noqa: C901
        self, message_name: str, payload: types.CommandPayload
    ) -> float:
        """
        Handle the message payload

        Args:
            message_name: Name of message
            payload: The command payload

        Returns:
            The duration in seconds
        """
        duration = 0.0
        # TODO (al, 2021-09-09):
        #  - Make this into a map
        #  - Remove "# noqa: C901"
        #  - type the payload in the on_X methods.
        if message_name == types.PICK_UP_TIP:
            duration = self.on_pick_up_tip(payload=payload)
        elif message_name == types.DROP_TIP:
            duration = self.on_drop_tip(payload=payload)
        elif message_name == types.ASPIRATE:
            duration = self.on_aspirate(payload=payload)
        elif message_name == types.DISPENSE:
            duration = self.on_dispense(payload=payload)
        elif message_name == types.BLOW_OUT:
            duration = self.on_blow_out(payload=payload)
        elif message_name == types.TOUCH_TIP:
            duration = self.on_touch_tip(payload=payload)
        elif message_name == types.DELAY:
            duration = self.on_delay(payload=payload)
        elif message_name == types.TEMPDECK_SET_TEMP:
            duration = self.on_tempdeck_set_temp(payload=payload)
        elif message_name == types.TEMPDECK_DEACTIVATE:
            duration = self.on_tempdeck_deactivate(payload=payload)
        elif message_name == types.TEMPDECK_AWAIT_TEMP:
            duration = self.on_tempdeck_await_temp(payload=payload)
        elif message_name == types.THERMOCYCLER_SET_BLOCK_TEMP:
            duration = self.on_thermocycler_block_temp(payload=payload)
        elif message_name == types.THERMOCYCLER_EXECUTE_PROFILE:
            duration = self.on_execute_profile(payload=payload)
        elif message_name == types.THERMOCYCLER_SET_LID_TEMP:
            duration = self.on_thermocycler_set_lid_temp(payload=payload)
        elif message_name == types.THERMOCYCLER_CLOSE:
            duration = self.on_thermocycler_lid_close(payload=payload)
        elif message_name == types.THERMOCYCLER_DEACTIVATE_LID:
            duration = self.on_thermocycler_deactivate_lid(payload=payload)
        elif message_name == types.THERMOCYCLER_OPEN:
            duration = self.on_thermocycler_lid_open(payload=payload)
        elif message_name == types.TRANSFER:
            # Already accounted for in other steps
            pass
        elif message_name == types.DISTRIBUTE:
            pass
        elif message_name == types.CONSOLIDATE:
            pass
        elif message_name == types.COMMENT:
            pass
        else:
            logger.warning(
                f"Command type '{message_name}' is not yet supported by the "
                f"duration estimator."
            )
        return duration

    def on_pick_up_tip(self, payload) -> float:
        """Handle a pick up tip event"""
        instrument = payload["instrument"]

        location = payload["location"]
        prev_slot = self._last_deckslot
        curr_slot = self.get_slot(location)

        gantry_speed = instrument.default_speed

        deck_travel_time = self.calc_deck_movement_time(
            self._deck, curr_slot, prev_slot, gantry_speed
        )

        # The time to pick up tip is 4 seconds. Determined by testing on hardware.
        duration = deck_travel_time + 4

        logger.info(
            f"{instrument.name} picked up tip from slot "
            f"{curr_slot} the duration is {duration}"
        )
        return duration

    def on_drop_tip(self, payload) -> float:
        instrument = payload["instrument"]
        # We are going to once again use our "deck movement" set up. This should
        # be in pickup, drop tip, aspirate, dispense
        location = payload["location"]
        curr_slot = self.get_slot(location)
        # this one disagrees with me.
        prev_slot = self._last_deckslot
        gantry_speed = instrument.default_speed
        deck_travel_time = self.calc_deck_movement_time(
            self._deck, curr_slot, prev_slot, gantry_speed
        )
        # TODO (al, 2021-09-08): Should we be checking for drop tip home_after = False?

        # The time to drop up tip is 10 seconds. Determined by testing on hardware.
        duration = deck_travel_time + 10

        # let's only log the message after the pick up tip is done.
        logger.info(f"{instrument.name}, drop tip duration is {duration}")
        return duration

    def on_aspirate(self, payload) -> float:
        # General aspiration code
        instrument = payload["instrument"]
        volume = payload["volume"]
        rate = payload["rate"] * instrument.flow_rate.aspirate

        aspiration_time = volume / rate

        # now lets handle the aspiration z-axis code.
        location = payload["location"]
        slot = self.get_slot(location)

        gantry_speed = instrument.default_speed
        z_total_time = self.z_time(
            location.labware.parent.parent.is_module, gantry_speed
        )

        # We are going to once again use our "deck movement" set up.
        # Might be changed in future PR if our travel
        # calculations adjust
        # This should be in pickup, drop tip, aspirate, dispense
        location = payload["location"]
        prev_slot = self._last_deckslot
        curr_slot = self.get_slot(location)

        gantry_speed = instrument.default_speed
        deck_travel_time = self.calc_deck_movement_time(
            self._deck, curr_slot, prev_slot, gantry_speed
        )
        duration = deck_travel_time + z_total_time + aspiration_time
        logger.info(
            f"{instrument.name} aspirate from {slot}, " f"the duration is {duration}"
        )
        return duration

    def on_dispense(self, payload) -> float:
        # General code for aspiration/dispense
        instrument = payload["instrument"]
        volume = payload["volume"]
        rate = payload["rate"] * instrument.flow_rate.dispense
        dispense_time = volume / rate

        # define variables
        location = payload["location"]
        slot = self.get_slot(location)
        gantry_speed = instrument.default_speed

        z_total_time = self.z_time(
            location.labware.parent.parent.is_module, gantry_speed
        )
        # We are going to once again use our "deck movement" set up.
        # This should be in pickup, drop tip, aspirate, dispense
        location = payload["location"]
        prev_slot = self._last_deckslot
        curr_slot = self.get_slot(location)

        gantry_speed = instrument.default_speed
        deck_travel_time = self.calc_deck_movement_time(
            self._deck, curr_slot, prev_slot, gantry_speed
        )

        duration = deck_travel_time + z_total_time + dispense_time

        logger.info(
            f"{instrument.name} dispensed from {slot}, the duration is {duration}"
        )
        return duration

    def on_blow_out(self, payload) -> float:
        location = payload["location"]
        curr_slot = self.get_slot(location)
        # In theory, we could use instrument.flow_rate.blow_out, but we don't
        # know how much is in the tip left to blow out
        # So we are defaulting to 0.5 seconds
        duration = 0.5
        logger.info(f"blowing_out_for {duration} seconds, in slot {curr_slot}")
        return duration

    def on_touch_tip(self, payload) -> float:
        # base assumption. Touch_tip takes 0.5 seconds This is consistent with a
        # ~7.5mm diameter (default 60mm/s, 4 sides)
        # plate = protocol.load_labware('corning_96_wellplate_360ul_flat', '1')
        # depth = plate['A1'].diameter
        # Then use the speed of the touch tip
        # ( plate = protocol.load_labware('corning_96_wellplate_360ul_flat', '1')
        # depth = plate['A1'].diameter
        duration = 0.5
        logger.info(f"touch_tip for {duration} seconds")
        return duration

    def on_delay(self, payload) -> float:
        # Explanation: we are gathering seconds and minutes here
        seconds_delay = payload["seconds"]
        minutes_delay = payload["minutes"]
        duration = seconds_delay + minutes_delay * 60
        # Note will need to multiply minutes by 60
        logger.info(f"delay for {seconds_delay} seconds and {minutes_delay} minutes")

        return duration

    def on_thermocycler_block_temp(self, payload) -> float:
        temperature = payload["temperature"]
        hold_time = payload["hold_time"]
        temp0 = self._last_thermocycler_module_temperature
        temp1 = temperature
        # we are referring to a thermocycler_handler(temp0, temp1) function.
        # Magic numbers come from testing and have been consistent
        temperature_changing_time = self.thermocycler_handler(temp0, temp1)
        if hold_time is None:
            hold_time = 0
        else:
            hold_time = float(hold_time)

        duration = temperature_changing_time + hold_time
        # Note will need to multiply minutes by 60
        logger.info(
            f"hold for {hold_time} seconds and set temp for {temperature}"
            f" C total duration {duration}"
        )

        return duration

    def on_execute_profile(self, payload) -> float:
        # Overview We need to run each time a temperature change happens
        # through thermocycler_handler and multiply
        # By the cycle count. Then we also (in parallel) do the same with delays

        profile_total_steps = payload["steps"]
        thermocycler_temperatures = [self._last_thermocycler_module_temperature]
        thermocycler_hold_times = []
        cycle_count = float(payload["text"].split(" ")[2])

        # We are going to need to treat this theromcycler part a bit differently
        # for a bit and just send out total times
        for step in profile_total_steps:
            thermocycler_temperatures.append(float(step["temperature"]))
            thermocycler_hold_times.append(float(step["hold_time_seconds"]))
        # Initializing variable
        total_hold_time = float(cycle_count) * float(sum(thermocycler_hold_times))
        # This takes care of the cumulative hold time
        # WE DON't Have a way to deal with this currently in the way we
        # have things set up.
        cycling_counter = []
        thermocycler_temperatures.pop(0)
        for thermocycler_counter in range(0, len(thermocycler_temperatures)):
            cycling_counter.append(
                self.thermocycler_handler(
                    float(thermocycler_temperatures[thermocycler_counter - 1]),
                    float(thermocycler_temperatures[thermocycler_counter]),
                )
            )

        # Sum hold time and cycling temp time
        duration = float(sum(cycling_counter) + total_hold_time)
        self._last_thermocycler_module_temperature = thermocycler_temperatures[-1]

        cycling_counter = []
        # Note will need to multiply minutes by 60
        logger.info(
            f"temperatures {sum(cycling_counter)}, "
            f"hold_times {total_hold_time} , cycles are {cycle_count}, "
            f"{duration}"
        )
        return duration

    def on_thermocycler_set_lid_temp(self, payload) -> float:
        # Hardware said ~1 minute
        duration = 60
        thermoaction = "set lid temperature"
        logger.info(f"thermocation =  {thermoaction}")
        return duration

    def on_thermocycler_lid_close(self, payload) -> float:
        # Hardware said ~24 seconds
        duration = 24
        thermoaction = "closing"
        logger.info(f"thermocation =  {thermoaction}")
        return duration

    def on_thermocycler_lid_open(self, payload) -> float:
        # Hardware said ~24 seconds
        duration = 24
        thermoaction = "opening"
        logger.info(f"thermocation =  {thermoaction}")
        return duration

    def on_thermocycler_deactivate_lid(self, payload) -> float:
        # Hardware said ~23 seconds
        duration = 23
        thermoaction = "Deactivating"
        logger.info(f"thermocation =  {thermoaction}")
        return duration

    def on_tempdeck_set_temp(self, payload) -> float:
        temperature_tempdeck = payload["celsius"]
        temp0 = self._last_temperature_module_temperature
        temp1 = float(temperature_tempdeck)
        duration = self.temperature_module(temp0, temp1)
        self._last_temperature_module_temperature = temp0
        logger.info(f"tempdeck {duration} ")
        return duration

    def thermocycler_handler(self, temp0: float, temp1: float) -> float:
        total = 0.0
        if temp1 - temp0 > 0:
            # heating up!
            if temp1 > THERMO_HIGH_THRESH:
                # the temp1 part that's over 70 is
                total = abs(temp1 - THERMO_HIGH_THRESH) / 2
                # the temp1 part that's under 70 is:
                total += abs(THERMO_HIGH_THRESH - temp0) / 4
            else:
                total = abs(temp1 - temp0) / 4
        # This is where the error is. if it's 10 and 94 this would not
        # @Matt please look into this
        elif temp1 - temp0 < 0:
            if temp1 >= THERMO_HIGH_THRESH:
                total = abs(temp1 - temp0) / 2
            elif temp1 >= THERMO_LOW_THRESH:
                total = abs(temp1 - temp0) / 1
            else:
                # 70 to 23 2 C/s
                total = abs(temp0 - THERMO_LOW_THRESH) / 0.5
                # 23 to temp1 0.1 C/s
                total += abs(temp1 - THERMO_LOW_THRESH) / 0.1

        return total

    def temperature_module(self, temp0: float, temp1: float) -> float:
        duration = 0.0
        if temp1 != temp0:
            if temp1 > TEMP_MOD_HIGH_THRESH:
                duration = self.rate_high(temp0, temp1)
            elif TEMP_MOD_LOW_THRESH <= temp1 <= TEMP_MOD_HIGH_THRESH:
                duration = self.rate_mid(temp0, temp1)
            elif temp1 < TEMP_MOD_LOW_THRESH:
                duration = self.rate_low(temp0, temp1)
        return duration

    def on_tempdeck_deactivate(self, payload) -> float:
        # TODO (al, 2021-09-08: Find an answer for this value.
        duration = 0.0
        logger.info("tempdeck deactivating")
        return duration

    def on_tempdeck_await_temp(self, payload) -> float:
        # The duration is accounted for in set temperature
        duration = 0.0
        logger.info("tempdeck awaiting temperature")
        return duration

    @staticmethod
    def get_slot(location) -> Optional[str]:
        """A utility function to extract the slot number from the location."""
        if isinstance(location, Location):
            return location.labware.first_parent()
        else:
            return LabwareLike(location).first_parent()

    @staticmethod
    def calc_deck_movement_time(
        deck: Deck, current_slot: Optional[str], previous_slot: str, gantry_speed: float
    ) -> float:
        # Quick summary we set coordinates for each deck slot and found ways
        # to move between deck slots.
        # Each deck slot is a key for a coordinate value
        # Moving between coordinate values
        if not current_slot:
            raise DurationEstimatorException(
                f"Current slot '{current_slot}' is not valid."
            )

        previous_deck_center = deck.position_for(previous_slot)
        if not previous_deck_center:
            raise DurationEstimatorException(
                f"Previous slot '{current_slot}' is not valid."
            )
        current_deck_center = deck.position_for(current_slot)

        x_difference = abs(current_deck_center.point.x - previous_deck_center.point.x)
        y_difference = abs(current_deck_center.point.y - previous_deck_center.point.y)

        if x_difference == 0 and y_difference == 0:
            # Inter slot movement defaults to half a second.
            deck_movement_time = 0.5
        else:
            deck_distance = math.sqrt((x_difference ** 2) + (y_difference ** 2))
            deck_movement_time = deck_distance / gantry_speed
        return deck_movement_time

    @staticmethod
    def z_time(is_module: bool, gantry_speed: float) -> float:
        # TODO (al, 2021-09-08): Use definitions from protocol context objects.
        z_default_labware_height = 177.8
        z_default_module_height = 95.25
        # 177.8 - 82.55 Where did we get 177.8 from?
        # Would it be better to just use
        # protocol_api.labware.Well.top
        # labware.top() ?

        if is_module:
            z_time = z_default_module_height / gantry_speed
        else:
            z_time = z_default_labware_height / gantry_speed

        return z_time

    @staticmethod
    def rate_high(temp0: float, temp1: float) -> float:
        """
        Calculate temp deck temperature change to a high temperature.

        Args:
            temp0: Current
            temp1: Target

        Returns:
            Duration in seconds.
        """
        val = 0.0
        if temp0 >= TEMP_MOD_HIGH_THRESH:
            val = abs(temp1 - temp0) / TEMP_MOD_RATE_HIGH_AND_ABOVE
        elif TEMP_MOD_LOW_THRESH < temp0 < TEMP_MOD_HIGH_THRESH:
            val = abs(temp0 - TEMP_MOD_HIGH_THRESH) / TEMP_MOD_RATE_LOW_TO_HIGH
            val += abs(temp1 - TEMP_MOD_HIGH_THRESH) / TEMP_MOD_RATE_HIGH_AND_ABOVE
        elif temp0 <= TEMP_MOD_LOW_THRESH:
            # the temp1 part that's under TEMP_MOD_HIGH_THRESH is:
            val = abs(TEMP_MOD_LOW_THRESH - temp0) / TEMP_MOD_RATE_ZERO_TO_LOW
            val += abs(TEMP_MOD_LOW_THRESH - temp1) / TEMP_MOD_RATE_HIGH_AND_ABOVE

        return val

    @staticmethod
    def rate_mid(temp0: float, temp1: float) -> float:
        """
        Calculate temp deck temperature change to a medium temperature.

        Args:
            temp0: Current
            temp1: Target

        Returns:
            Duration in seconds.
        """
        val = 0.0
        if TEMP_MOD_LOW_THRESH <= temp0 <= TEMP_MOD_HIGH_THRESH:
            val = abs(temp1 - temp0) / TEMP_MOD_RATE_LOW_TO_HIGH
        elif temp0 < TEMP_MOD_LOW_THRESH:
            # the temp1 part that's over TEMP_MOD_HIGH_THRESH is
            val = abs(temp1 - TEMP_MOD_LOW_THRESH) / TEMP_MOD_RATE_LOW_TO_HIGH
            # the temp0 part that's under TEMP_MOD_HIGH_THRESH is:
            val += abs(TEMP_MOD_LOW_THRESH - temp0) / TEMP_MOD_RATE_ZERO_TO_LOW
        elif temp0 > TEMP_MOD_HIGH_THRESH:
            val = abs(temp0 - TEMP_MOD_HIGH_THRESH) / TEMP_MOD_RATE_HIGH_AND_ABOVE
            # the temp1 part that's under TEMP_MOD_HIGH_THRESH is:
            val += abs(TEMP_MOD_HIGH_THRESH - temp1) / TEMP_MOD_RATE_ZERO_TO_LOW
        return val

    # How to handle temperatures where one of them is low temp
    @staticmethod
    def rate_low(temp0: float, temp1: float) -> float:
        """
        Calculate temp deck temperature change to a low temperature.

        Args:
            temp0: Current
            temp1: Target

        Returns:
            Duration in seconds.
        """
        val = 0.0
        if temp0 <= TEMP_MOD_LOW_THRESH:
            val = abs(temp1 - temp0) / TEMP_MOD_RATE_ZERO_TO_LOW
        elif TEMP_MOD_LOW_THRESH < temp0 < TEMP_MOD_HIGH_THRESH:
            # the temp0 part that's over TEMP_MOD_HIGH_THRESH is
            val = abs(temp0 - TEMP_MOD_LOW_THRESH) / TEMP_MOD_RATE_LOW_TO_HIGH
            # the temp1 part that's under TEMP_MOD_HIGH_THRESH is:
            val += abs(TEMP_MOD_LOW_THRESH - temp1) / TEMP_MOD_RATE_ZERO_TO_LOW
        elif temp0 >= TEMP_MOD_HIGH_THRESH:
            val = abs(temp0 - TEMP_MOD_HIGH_THRESH) / TEMP_MOD_RATE_LOW_TO_HIGH
            # the temp1 part that's under TEMP_MOD_HIGH_THRESH is:
            val += abs(TEMP_MOD_HIGH_THRESH - temp1) / TEMP_MOD_RATE_ZERO_TO_LOW
        return val



from __future__ import annotations
import asyncio
import base64
from copy import copy
from functools import reduce
import logging
from time import time, sleep
from typing import (
    cast,
    List,
    Dict,
    Any,
    Optional,
    Set,
    Sequence,
    Tuple,
    TypeVar,
    Iterator,
)
from typing_extensions import Final
from uuid import uuid4

from opentrons.types import Mount as MountType
from opentrons_shared_data.labware.dev_types import LabwareDefinition

from opentrons.api.util import (
    RobotBusy,
    robot_is_busy,
    requires_http_protocols_disabled,
)
from opentrons.drivers.smoothie_drivers.errors import SmoothieAlarm
from opentrons.broker import Broker
from opentrons.config import feature_flags as ff
from opentrons.commands.util import from_list
from opentrons.commands import types as command_types, introspection
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.context.protocol_api.protocol_context import (
    ProtocolContextImplementation,
)
from opentrons.protocols.context.simulator.protocol_context import (
    ProtocolContextSimulation,
)
from opentrons.protocols.parse import parse
from opentrons.protocols.types import Protocol
from opentrons.calibration_storage import helpers
from opentrons.protocol_api import ProtocolContext, InstrumentContext, labware
from opentrons.protocols.geometry import module_geometry
from opentrons.protocols.execution.execute import run_protocol
from opentrons.hardware_control import (
    API,
    ThreadManager,
    SynchronousAdapter,
    ExecutionCancelledError,
    ThreadedAsyncLock,
)
from opentrons.hardware_control.types import Axis
from opentrons.hardware_control.types import HardwareEventType, HardwareEvent, PauseType
from .models import Container, Instrument, Module
from .dev_types import State, StateInfo, Message, LastCommand, Error, CommandShortId

log = logging.getLogger(__name__)

VALID_STATES: Set[State] = {
    "loaded",
    "running",
    "finished",
    "stopped",
    "paused",
    "error",
}


class SessionManager:
    def __init__(
        self,
        hardware: SynchronousAdapter,
        loop: asyncio.AbstractEventLoop = None,
        broker: Broker = None,
        lock: ThreadedAsyncLock = None,
    ) -> None:
        self._broker = broker or Broker()
        self._loop = loop or asyncio.get_event_loop()
        self.session: Optional[Session] = None
        self._session_lock = False
        self._hardware = hardware
        self._command_logger = logging.getLogger("opentrons.server.command_logger")
        self._broker.set_logger(self._command_logger)
        self._motion_lock = lock or ThreadedAsyncLock()

    @requires_http_protocols_disabled
    def create(self, name: str, contents: str, is_binary: bool = False) -> Session:
        """Create a protocol session from either

        - a json protocol
        - a python protocol file
        - a zipped protocol bundle (deprecated, for back compat)

        No new code should be written that calls this function with
        ``is_binary=True`` and a base64'd zip in ``contents``; instead,
        use :py:meth:`.create_from_bundle`.

        :param str name: The name of the protocol
        :param str contents: The contents of the protocol; this should be
                             either the contents of the file if it can be
                             parsed directly or a base64d version of the
                             contents if it is a zip. If it is base64,
                             ``is_binary`` must be true. Do not write new
                             code that uses this.
        :param bool is_binary: ``True`` if ``contents`` is a base64'd zip.
                               Do not write new code that uses this.
        :returns Session: The created session.
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if is_binary:
            log.warning("session.create: called with bundle")

        if self._session_lock:
            raise Exception("Cannot create session while simulation in progress")

        self.clear()
        self._session_lock = True
        try:
            _contents = base64.b64decode(contents) if is_binary else contents
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=_contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=[],
            )
            return self.session
        finally:
            self._session_lock = False

    @requires_http_protocols_disabled
    def create_from_bundle(self, name: str, contents: str) -> Session:
        """Create a protocol session from a base64'd zip file.

        :param str name: The name of the protocol
        :param str contents: The contents of the zip file, base64
                             encoded
        :returns Session: The created session
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if self._session_lock:
            raise Exception("Cannot create session while simulation in progress")

        self.clear()
        self._session_lock = True
        try:
            _contents = base64.b64decode(contents)
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=_contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=[],
            )
            return self.session
        finally:
            self._session_lock = False

    @requires_http_protocols_disabled
    def create_with_extra_labware(
        self, name: str, contents: str, extra_labware: List[LabwareDefinition]
    ) -> Session:
        """
        Create a protocol session from a python protocol string with a list
        of extra custom labware to make available.

        :param str name: The name of the protocol
        :param str contents: The contents of the protocol file
        :param extra_labware: A list of labware definitions to make available
                              to protocol executions. This should be a list
                              of directly serialized definitions.
        :returns Session: The created session
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if self._session_lock:
            raise Exception("Cannot create session while simulation in progress")

        self.clear()
        self._session_lock = True
        try:
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=extra_labware,
            )
            return self.session
        finally:
            self._session_lock = False

    def clear(self) -> None:
        if self._session_lock:
            raise Exception("Cannot clear session while simulation in progress")

        if self.session:
            self.session._remove_hardware_event_watcher()
            self._hardware.reset()
        self.session = None
        self._broker.set_logger(self._command_logger)

    def get_session(self) -> Optional[Session]:
        return self.session


class Session(RobotBusy):
    TOPIC: Final = "session"

    @classmethod
    def build_and_prep(
        cls,
        name: str,
        contents: Any,
        hardware: SynchronousAdapter,
        loop: asyncio.AbstractEventLoop,
        broker: Broker,
        motion_lock: ThreadedAsyncLock,
        extra_labware: List[LabwareDefinition],
    ) -> Session:
        protocol = parse(
            contents,
            filename=name,
            extra_labware={
                helpers.uri_from_definition(defn): defn for defn in extra_labware
            },
        )
        sess = cls(name, protocol, hardware, loop, broker, motion_lock)
        sess.prepare()
        return sess

    def __init__(
        self,
        name: str,
        protocol: Protocol,
        hardware: SynchronousAdapter,
        loop: asyncio.AbstractEventLoop,
        broker: Broker,
        motion_lock: ThreadedAsyncLock,
    ) -> None:
        self._broker = broker
        self._default_logger = self._broker.logger
        self._sim_logger = self._broker.logger.getChild("sim")
        self._run_logger = self._broker.logger.getChild("run")
        self._loop = loop
        self.name = name
        self._protocol = protocol
        self.api_level = getattr(self._protocol, "api_level", APIVersion(2, 0))
        self._use_v2 = self.api_level >= APIVersion(2, 0)
        log.info(
            f"Protocol API Version: {self.api_level}; "
            f"Protocol kind: {type(self._protocol)}"
        )

        # self.metadata is exposed via rpc
        self.metadata = getattr(self._protocol, "metadata", {})

        self._hardware = hardware
        self._simulating_ctx = ProtocolContext.build_using(
            implementation=ProtocolContextImplementation.build_using(self._protocol),
            protocol=self._protocol,
            loop=self._loop,
            broker=self._broker,
        )

        self.state: "State" = None
        #: The current state
        self.stateInfo: "StateInfo" = {}
        #: A message associated with the current state
        self.commands: List[command_types.CommandMessage] = []
        self.command_log: Dict[int, int] = {}
        self.errors: List[Error] = []

        # Underlying objects harvested from commands, internal
        self._containers: List[labware.Labware] = []
        self._instruments: List[InstrumentContext] = []
        self._modules: List[module_geometry.ModuleGeometry] = []
        self._interactions: List[Tuple[InstrumentContext, labware.Labware]] = []

        # RPC-safe models of objects harvested from commands
        self.instruments: Optional[List[Instrument]] = None
        self.containers: Optional[List[Container]] = None
        self.modules: Optional[List[Module]] = None
        self.protocol_text = protocol.text

        self.startTime: Optional[float] = None
        self._motion_lock = motion_lock
        self._event_watcher = None
        self.door_state: Optional[str] = None
        self.blocked: Optional[bool] = None

    @property
    def busy_lock(self) -> ThreadedAsyncLock:
        return self._motion_lock

    def _hw_iface(self) -> SynchronousAdapter:
        return self._hardware

    def prepare(self) -> None:
        self.refresh()

    def get_instruments(self) -> List[Instrument]:
        return [
            Instrument(
                instrument=instrument,
                containers=[
                    container
                    for _instrument, container in self._interactions
                    if _instrument == instrument
                ],
                context=self._use_v2 and self._simulating_ctx,
            )
            for instrument in self._instruments
        ]

    def get_containers(self) -> List[Container]:
        return [
            Container(
                container=container,
                instruments=[
                    instrument
                    for instrument, _container in self._interactions
                    if _container == container
                ],
                context=self._use_v2 and self._simulating_ctx,
            )
            for container in self._containers
        ]

    def get_modules(self) -> List[Module]:
        return [
            Module(module=module, context=self._use_v2 and self._simulating_ctx)
            for module in self._modules
        ]

    def clear_logs(self) -> None:
        self.command_log.clear()
        self.errors.clear()

    @robot_is_busy
    def _simulate(self) -> List[CommandShortId]:
        self._reset()

        stack: List[command_types.CommandMessage] = []
        res: List[CommandShortId] = []
        commands: List[command_types.CommandPayload] = []

        self._containers.clear()
        self._instruments.clear()
        self._modules.clear()
        self._interactions.clear()

        def on_command(message: command_types.CommandMessage) -> None:
            payload = message["payload"]
            description = payload.get("text", "").format(**payload)

            if message["$"] == "before":
                level = len(stack)

                stack.append(message)
                commands.append(payload)

                res.append({"level": level, "description": description, "id": len(res)})
            else:
                stack.pop()

        unsubscribe = self._broker.subscribe(command_types.COMMAND, on_command)

        try:
            # ensure actual pipettes are cached before driver is disconnected
            self._hardware.cache_instruments()
            instrs = {}
            for mount, pip in self._hardware.attached_instruments.items():
                if pip:
                    instrs[mount] = {
                        "model": pip["model"],
                        "id": pip.get("pipette_id", ""),
                    }
            sync_sim = ThreadManager(
                API.build_hardware_simulator,
                instrs,
                [mod.name() for mod in self._hardware.attached_modules],
                strict_attached_instruments=False,
            ).sync
            sync_sim.home()

            # Use feature flag to determine ProtocolContext implementation
            # class: real or simulating.
            impl_class = (
                ProtocolContextImplementation
                if ff.disable_fast_protocol_upload()
                else ProtocolContextSimulation
            )

            ctx_impl = impl_class.build_using(self._protocol, hardware=sync_sim)
            self._simulating_ctx = ProtocolContext.build_using(
                protocol=self._protocol,
                loop=self._loop,
                broker=self._broker,
                implementation=ctx_impl,
            )
            run_protocol(self._protocol, context=self._simulating_ctx)

        finally:
            unsubscribe()

            instruments, containers, modules, interactions = _accumulate(
                [_get_labware(command) for command in commands]
            )

            self._containers.extend(_dedupe(containers))
            self._instruments.extend(
                _dedupe(
                    instruments + list(self._simulating_ctx.loaded_instruments.values())
                )
            )
            self._modules.extend(
                _dedupe(
                    modules
                    + [
                        m._geometry
                        for m in self._simulating_ctx.loaded_modules.values()
                    ]
                )
            )
            self._interactions.extend(_dedupe(interactions))

            # Labware calibration happens after simulation and before run, so
            # we have to clear the tips if they are left on after simulation
            # to ensure that the instruments are in the expected state at the
            # beginning of the labware calibration flow

        return res

    def refresh(self) -> None:
        self._reset()

        try:
            self._broker.set_logger(self._sim_logger)
            commands = self._simulate()
        except Exception:
            raise
        finally:
            self._broker.set_logger(self._default_logger)

        self.commands = from_list(commands)

        self.containers = self.get_containers()
        self.instruments = self.get_instruments()
        self.modules = self.get_modules()
        self.startTime = None
        self.set_state("loaded")

    def stop(self) -> None:
        # Save the current pipette tip length before calling self._hardware.halt
        # so we don't lose it on hard halt. Used for determining a need for tip dropping
        instruments = self._hardware.attached_instruments
        saved_tip_length = {
            mount: pip.get("tip_length") for mount, pip in instruments.items()
        }
        self._hardware.halt()
        with self._motion_lock.lock():
            try:
                if any(saved_tip_length.values()):
                    self._hardware.stop(home_after=False)
                    self._hardware.home(axes=Axis.gantry_axes())
                    self._drop_tip_after_cancel(saved_tip_length)
                self._hardware.stop(home_after=True)
            except asyncio.CancelledError:
                pass
        self.set_state("stopped")

    def _drop_tip_after_cancel(
        self,
        saved_tip_length: Dict[MountType, Optional[float]],
    ) -> None:
        """Perform a drop_tip on the pipette with a tip.

        Note: Moving the plunger the right amount to drop the tip depends on
              knowing the previous plunger location. When the smoothie board is issued
              a `halt`, we need to reset it in order to resume communication.
              BUT, it seems like the smoothie doesn't wipe out the cached locations upon
              reset, so, we can just resume moving the plunger without homing it first.
              This behavior has been consistent so far but if ends up being unreliable
              then we will have to change this logic.

              Also, `self._hardware.halt()` does not stop the motors gracefully. If the
              plunger motor was moving at a speed above the pull-in speed at the time,
              it could skip. So, even if the motor controller board's memory of the
              plunger's current location isn't wiped out, it could be slightly wrong.
              That said, since plunger speeds during liquid handling commands are much
              lower, the probability of this causing an issue is very low.
        """
        if not self.instruments:
            return
        for instrument_wrapper in self.instruments:
            mount = MountType.string_to_mount(instrument_wrapper.mount)
            protocol_ctx = instrument_wrapper._context
            tip_length = saved_tip_length[mount]

            if tip_length:
                self._hardware.add_tip(mount=mount, tip_length=tip_length)
                with protocol_ctx.temp_connect(self._hardware):
                    protocol_ctx.location_cache = None
                    instrument_wrapper._instrument.drop_tip()

    def pause(
        self, reason: str = None, user_message: str = None, duration: float = None
    ) -> None:
        self._hardware.pause(PauseType.PAUSE)
        self.set_state(
            "paused", reason=reason, user_message=user_message, duration=duration
        )

    def resume(self) -> None:
        if not self.blocked:
            self._hardware.resume(PauseType.PAUSE)
            self.set_state("running")

    def _start_hardware_event_watcher(self) -> None:
        if not callable(self._event_watcher):
            # initialize and update window switch state
            self.door_state = str(self._hardware.door_state)
            self.blocked = self._hardware._pause_manager.blocked_by_door
            log.info("Starting hardware event watcher")
            self._event_watcher = self._hardware.register_callback(
                self._handle_hardware_event
            )
        else:
            log.warning(
                "Cannot start new hardware event watcher " "when one already exists"
            )

    def _remove_hardware_event_watcher(self) -> None:
        if self._event_watcher and callable(self._event_watcher):
            self._event_watcher()
            self._event_watcher = None

    def _handle_hardware_event(self, hw_event: HardwareEvent) -> None:
        if hw_event.event == HardwareEventType.DOOR_SWITCH_CHANGE:
            self.door_state = str(hw_event.new_state)
            self.blocked = hw_event.blocking
            if self.blocked and self.state == "running":
                self.pause("Robot door is open")
            else:
                self._on_state_changed()

    @robot_is_busy
    def _run(self) -> None:
        def on_command(message: command_types.CommandMessage) -> None:
            if message["$"] == "before":
                self.log_append()
            if message["name"] == command_types.PAUSE:
                self.set_state(
                    "paused",
                    reason="The protocol paused execution",
                    user_message=message["payload"]["userMessage"],
                )
            if message["name"] == command_types.RESUME:
                self.set_state("running")

        self._reset()

        _unsubscribe = self._broker.subscribe(command_types.COMMAND, on_command)

        self.startTime = now()
        self.set_state("running")

        try:
            self.resume()
            self._pre_run_hooks()
            self._hardware.cache_instruments()
            self._hardware.reset_instrument()
            ctx_impl = ProtocolContextImplementation.build_using(
                self._protocol,
                extra_labware=getattr(self._protocol, "extra_labware", {}),
            )
            ctx = ProtocolContext.build_using(
                protocol=self._protocol,
                implementation=ctx_impl,
                loop=self._loop,
                broker=self._broker,
            )
            ctx.connect(self._hardware)
            ctx.home()
            run_protocol(self._protocol, context=ctx)

            # If the last command in a protocol was a pause, the protocol
            # will immediately finish executing because there's no smoothie
            # command to block... except the home that's about to happen,
            # which will confuse the app and lock it up. So we need to
            # do our own pause here, and sleep the thread until/unless the
            # app resumes us.
            #
            # Cancelling from the app during this pause will result in the
            # smoothie giving us an error during the subsequent home, which
            # is tragic but expected.
            while self.state == "paused":
                sleep(0.1)
            self.set_state("finished")
            self._hw_iface().home()
        except (SmoothieAlarm, asyncio.CancelledError, ExecutionCancelledError):
            log.info("Protocol cancelled")
        except Exception as e:
            log.exception("Exception during run:")
            self.error_append(e)
            self.set_state("error")
            raise e
        finally:
            _unsubscribe()

    def run(self) -> None:
        if not self.blocked:
            try:
                self._broker.set_logger(self._run_logger)
                self._run()
            except Exception:
                raise
            finally:
                self._broker.set_logger(self._default_logger)
        else:
            raise RuntimeError(
                "Protocol is blocked and cannot run. Make sure robot door "
                "is closed before running."
            )

    def set_state(
        self,
        state: "State",
        reason: str = None,
        user_message: str = None,
        duration: float = None,
    ) -> None:
        log.info(
            f"Session.set_state state={state} "
            f"user_message={user_message} duration={duration} "
            f"reason={reason}"
        )
        if state not in VALID_STATES:
            raise ValueError(
                "Invalid state: {0}. Valid states are: {1}".format(state, VALID_STATES)
            )
        self.state = state
        if user_message:
            self.stateInfo["userMessage"] = user_message
        else:
            self.stateInfo.pop("userMessage", None)
        if reason:
            self.stateInfo["message"] = reason
        else:
            self.stateInfo.pop("message", None)
        if duration:
            self.stateInfo["estimatedDuration"] = duration
        else:
            self.stateInfo.pop("estimatedDuration", None)
        if self.startTime:
            self.stateInfo["changedAt"] = now() - self.startTime
        else:
            self.stateInfo.pop("changedAt", None)
        self._on_state_changed()

    def log_append(self) -> None:
        self.command_log.update({len(self.command_log): now()})
        self._on_state_changed()

    def error_append(self, error: Exception) -> None:
        self.errors.append({"timestamp": now(), "error": error})

    def _reset(self) -> None:
        self._hw_iface().reset()
        self.clear_logs()
        # unregister existing event watcher
        self._remove_hardware_event_watcher()
        self._start_hardware_event_watcher()

    def _snapshot(self) -> Message:
        if self.state == "loaded":
            payload: Any = copy(self)
        else:
            if self.command_log.keys():
                idx = sorted(self.command_log.keys())[-1]
                timestamp = self.command_log[idx]
                last_command: Optional[LastCommand] = {
                    "id": idx,
                    "handledAt": timestamp,
                }
            else:
                last_command = None

            payload = {
                "state": self.state,
                "stateInfo": self.stateInfo,
                "startTime": self.startTime,
                "doorState": self.door_state,
                "blocked": self.blocked,
                "errors": self.errors,
                "lastCommand": last_command,
            }
        return {"topic": Session.TOPIC, "payload": payload}

    def _on_state_changed(self) -> None:
        snap = self._snapshot()
        self._broker.publish(Session.TOPIC, snap)

    def _pre_run_hooks(self) -> None:
        self._hw_iface().home_z()


CommandReferents = Tuple[
    List[InstrumentContext],
    List[labware.Labware],
    List[module_geometry.ModuleGeometry],
    List[Tuple[InstrumentContext, labware.Labware]],
]


def _accumulate(iterable: Sequence[CommandReferents]) -> CommandReferents:
    def _reducer(acc: CommandReferents, item: CommandReferents) -> CommandReferents:
        return (acc[0] + item[0], acc[1] + item[1], acc[2] + item[2], acc[3] + item[3])

    return reduce(_reducer, iterable, cast(CommandReferents, ([], [], [], [])))


It = TypeVar("It")


def _dedupe(iterable: Sequence[It]) -> Iterator[It]:
    acc: Set[It] = set()

    for item in iterable:
        if item not in acc:
            acc.add(item)
            yield item


def now() -> int:
    return int(time() * 1000)


def _get_labware(command: command_types.CommandPayload) -> CommandReferents:
    interactions: List[Tuple[InstrumentContext, labware.Labware]] = []

    try:
        instruments, containers, modules = introspection.get_referred_objects(command)
    except ValueError:
        log.exception(f"Cant handle location in command {command!r}")
        instruments = []
        containers = []
        modules = []

    for instrument in instruments:
        interactions.extend([(instrument, container) for container in containers])

    return instruments, containers, modules, interactions



"""Command side-effect execution logic container."""
from logging import getLogger
from typing import Optional

from ..state import StateStore, UpdateCommandAction
from ..resources import ModelUtils
from ..commands import CommandStatus, CommandMapper
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .pipetting import PipettingHandler
from .run_control import RunControlHandler

log = getLogger(__name__)


class CommandExecutor:
    """CommandExecutor container class.

    CommandExecutor manages various child handlers that define procedures to
    execute the side-effects of commands.
    """

    def __init__(
        self,
        state_store: StateStore,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        run_control: RunControlHandler,
        command_mapper: Optional[CommandMapper] = None,
        model_utils: Optional[ModelUtils] = None,
    ) -> None:
        """Initialize the CommandExecutor with access to its dependencies."""
        self._state_store = state_store
        self._equipment = equipment
        self._movement = movement
        self._pipetting = pipetting
        self._run_control = run_control
        self._command_mapper = command_mapper or CommandMapper()
        self._model_utils = model_utils or ModelUtils()

    async def execute(self, command_id: str) -> None:
        """Run a given command's execution procedure.

        Arguments:
            command_id: The identifier of the command to execute. The
                command itself will be looked up from state.
        """
        command = self._state_store.commands.get(command_id=command_id)
        command_impl = command._ImplementationCls(
            equipment=self._equipment,
            movement=self._movement,
            pipetting=self._pipetting,
            run_control=self._run_control,
        )

        started_at = self._model_utils.get_timestamp()
        running_command = self._command_mapper.update_command(
            command=command,
            status=CommandStatus.RUNNING,
            startedAt=started_at,
        )

        self._state_store.handle_action(UpdateCommandAction(command=running_command))

        result = None
        error = None
        try:
            log.debug(f"Executing {command.id}, {command.commandType}, {command.data}")
            result = await command_impl.execute(command.data)  # type: ignore[arg-type]
            completed_status = CommandStatus.SUCCEEDED
        except Exception as e:
            log.warn(
                f"Execution of {command.id} failed",
                exc_info=e,
            )
            # TODO(mc, 2021-06-22): differentiate between `ProtocolEngineError`s
            # and unexpected errors when the Command model is ready to accept
            # structured error details
            error = str(e)
            completed_status = CommandStatus.FAILED

        completed_at = self._model_utils.get_timestamp()
        completed_command = self._command_mapper.update_command(
            command=running_command,
            result=result,
            error=error,
            status=completed_status,
            completedAt=completed_at,
        )

        self._state_store.handle_action(UpdateCommandAction(command=completed_command))



from typing import List

from opentrons.config import feature_flags as ff

from .types import DoorState, PauseType, PauseResumeError


class PauseManager:
    """This class determines whether or not the hardware controller should
    pause or resume by evaluating the pause and resume types. The use of two
    pause types are used to separate the delay resume (triggered when the delay
    timer runs out) and the pause resume (trigged by user via the app).
    """

    def __init__(self, door_state: DoorState) -> None:
        self.queue: List[PauseType] = []
        self._blocked_by_door = self._evaluate_door_state(door_state)

    @property
    def should_pause(self) -> bool:
        return bool(self.queue)

    @property
    def blocked_by_door(self) -> bool:
        return self._blocked_by_door

    def _evaluate_door_state(self, door_state: DoorState) -> bool:
        if ff.enable_door_safety_switch():
            return door_state is DoorState.OPEN
        return False

    def set_door(self, door_state: DoorState):
        self._blocked_by_door = self._evaluate_door_state(door_state)

    def resume(self, pause_type: PauseType):
        # door should be closed before a resume from the app can be received
        if self._blocked_by_door and pause_type is PauseType.PAUSE:
            raise PauseResumeError
        if pause_type in self.queue:
            self.queue.remove(pause_type)

    def pause(self, pause_type: PauseType):
        if pause_type not in self.queue:
            self.queue.append(pause_type)

    def reset(self):
        self.queue = []



# Convenience re-exports of models that are especially common or important.
# More detailed sub-models are always available through the underlying
# submodules.
#
# If re-exporting something, its name should still make sense when it's separatred
# from the name of its parent submodule. e.g. re-exporting models.json_protocol.Labware
# as models.Labware could be confusing.

from .labware_definition import LabwareDefinition, WellDefinition
from .json_protocol import Model as JsonProtocol

__all__ = [
    "LabwareDefinition",
    "WellDefinition",
    "JsonProtocol",
]



import asyncio
from abc import abstractmethod, ABC
from collections import deque
from typing import TypeVar, Generic, Deque, Optional

DataT = TypeVar("DataT")


class Reader(ABC, Generic[DataT]):
    """Interface of poller target."""

    @abstractmethod
    async def read(self) -> DataT:
        """
        Read a new poll sample.

        Returns: The next poll result.
        """
        ...


class Listener(ABC, Generic[DataT]):
    """Interface of poller listener"""

    @abstractmethod
    def on_poll(self, result: DataT) -> None:
        """
        Called by poller notifying result of new poll.

        Args:
            result: The latest poll result.

        Returns: None
        """
        ...

    @abstractmethod
    def on_error(self, exception: Exception) -> None:
        """
        Called by poller to notify of a poll error.

        Args:
            exception: The raised exception

        Returns: None
        """
        ...

    @abstractmethod
    def on_terminated(self) -> None:
        """
        Called by poller when it is terminating.

        Returns: None
        """
        ...


class WaitableListener(Listener[DataT]):
    """A listener that can be waited on."""

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """Constructor."""
        self._loop = loop or asyncio.get_running_loop()
        self._futures: Deque[asyncio.Future] = deque()

    async def wait_next_poll(self) -> DataT:
        """
        Wait for the next poll.

        Returns: The next poll result.
        """
        f = self._loop.create_future()
        self._futures.append(f)
        return await f

    def on_poll(self, result: DataT) -> None:
        """Handle a new poll"""
        self._notify(result)

    def on_error(self, exc: Exception) -> None:
        """Handle a poller error."""
        self._cancel(exc)

    def on_terminated(self) -> None:
        """Handle the poller finishing up."""
        self._cancel(Exception("Poller has terminated."))

    def _notify(self, val: DataT) -> None:
        """Notify all futures of a new poll result."""
        while self._futures:
            f = self._futures.popleft()
            f.set_result(val)

    def _cancel(self, exc: Exception) -> None:
        """Notify all futures of an error."""
        while self._futures:
            f = self._futures.popleft()
            f.set_exception(exc)


class Poller(Generic[DataT]):
    """Asyncio poller."""

    def __init__(
        self,
        interval_seconds: float,
        reader: Reader[DataT],
        listener: Listener[DataT],
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """
        Constructor.

        Args:
            interval_seconds: time in between polls.
            reader: The data reader.
            listener: event listener.
            loop: Optional event loop to use
        """
        loop = loop or asyncio.get_running_loop()
        self._shutdown_event = asyncio.Event(loop=loop)
        self._interval = interval_seconds
        self._listener = listener
        self._reader = reader
        self._task = loop.create_task(self._poller())

    def stop(self) -> None:
        """Signal poller to stop."""
        self._shutdown_event.set()

    async def stop_and_wait(self) -> None:
        """Stop poller and wait for it to terminate."""
        self.stop()
        await self._task

    async def _poller(self) -> None:
        """Poll task entrypoint."""
        while True:
            try:
                poll = await self._reader.read()
                self._listener.on_poll(poll)
            except Exception as e:
                self._listener.on_error(e)

            try:
                await asyncio.wait_for(self._shutdown_event.wait(), self._interval)
            except asyncio.TimeoutError:
                pass
            else:
                break

        self._listener.on_terminated()



from __future__ import annotations

from typing import cast, Tuple, TYPE_CHECKING, List, Union, Optional
from opentrons.types import Location
from opentrons.protocols.api_support import labware_like

import opentrons.commands.types as command_types

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Labware, Well
    from opentrons.protocols.geometry.module_geometry import ModuleGeometry

    ReferredObjects = Tuple[
        List[InstrumentContext], List[Labware], List[ModuleGeometry]
    ]


def _labware_from_location(loc: Union[Location, Well]) -> Labware:
    if isinstance(loc, Location):
        if not loc.labware:
            raise ValueError("Cannot handle location without labware")
        labware, _ = loc.labware.get_parent_labware_and_well()
        if not labware:
            raise ValueError(f"Cannot handle location {loc}")
        return labware
    else:
        return loc.parent


# The casts and type: ignores in _get_locations and _get_instruments are necessary
# until we either a) refactor this so payloads have tags that mypy can use to
# discriminate the unions (or make commands dataclasses that can be checked with
# isinstance) or b) mypy learns full disjoint union discrimination.


def _get_locations(
    command: command_types.HasLocationPayload,
) -> List[Optional[Labware]]:
    if "location" in command and command["location"]:  # type: ignore
        return [
            _labware_from_location(
                cast(command_types.SingleLocationPayload, command)["location"]
            )
        ]
    elif "locations" in command and command["locations"]:  # type: ignore
        return [
            _labware_from_location(loc)
            for loc in cast(command_types.MultiLocationPayload, command)["locations"]
        ]
    else:
        return []


def _get_instruments(
    command: command_types.HasInstrumentPayload,
) -> List[InstrumentContext]:
    if "instrument" in command:
        return [cast(command_types.SingleInstrumentPayload, command)["instrument"]]
    else:
        return [
            inst
            for inst in cast(command_types.MultiInstrumentPayload, command)[
                "instruments"
            ]
        ]


def get_referred_objects(command: command_types.CommandPayload) -> ReferredObjects:
    if "location" in command or "locations" in command:
        locations = [
            lw
            for lw in _get_locations(cast(command_types.HasLocationPayload, command))
            if lw is not None
        ]
    else:
        locations = []
    if "instrument" in command or "instruments" in command:
        instruments = _get_instruments(
            cast(command_types.HasInstrumentPayload, command)
        )
    else:
        instruments = []

    maybe_modules = [
        labware_like.LabwareLike(labware).module_parent() for labware in locations
    ]
    return instruments, locations, [mod for mod in maybe_modules if mod]



from enum import Enum, auto
from typing import TYPE_CHECKING, Optional, Union, cast, Tuple, List, Set

if TYPE_CHECKING:
    from opentrons.protocol_api.labware import Labware, Well
    from opentrons.protocols.geometry.module_geometry import ModuleGeometry


WrappableLabwareLike = Union[
    "Labware", "Well", str, "ModuleGeometry", "LabwareLike", None
]


class LabwareLikeType(int, Enum):
    LABWARE = auto()
    SLOT = auto()
    WELL = auto()
    MODULE = auto()
    NONE = auto()


class LabwareLike:
    """A wrapper for a labware like object."""

    def __init__(self, labware_like: WrappableLabwareLike):
        """
        Create a labware like object. Used by Location object's labware field.
        """
        # Import locally to avoid circular dependency
        from opentrons.protocol_api.labware import Labware, Well
        from opentrons.protocols.geometry.module_geometry import ModuleGeometry

        self._labware_like = labware_like
        self._type = LabwareLikeType.NONE

        if isinstance(self._labware_like, Well):
            self._type = LabwareLikeType.WELL
            self._as_str = repr(self._labware_like)
        elif isinstance(self._labware_like, Labware):
            self._type = LabwareLikeType.LABWARE
            self._as_str = repr(self._labware_like)
        elif isinstance(self._labware_like, str):
            self._type = LabwareLikeType.SLOT
            self._as_str = self._labware_like
        elif isinstance(self._labware_like, ModuleGeometry):
            self._type = LabwareLikeType.MODULE
            self._as_str = repr(self._labware_like)
        elif isinstance(self._labware_like, LabwareLike):
            self._type = self._labware_like._type
            self._as_str = self._labware_like._as_str
            self._labware_like = self._labware_like.object
        else:
            self._as_str = ""

    @property
    def object(self) -> WrappableLabwareLike:
        return self._labware_like

    @property
    def object_type(self) -> LabwareLikeType:
        return self._type

    @property
    def has_parent(self) -> bool:
        return self._type in {
            LabwareLikeType.LABWARE,
            LabwareLikeType.WELL,
            LabwareLikeType.MODULE,
        }

    @property
    def parent(self) -> "LabwareLike":
        # Type ignoring because type checker is not aware that has_parent
        # performs the validation.
        parent = None
        if self.has_parent:
            parent = self.object.parent  # type: ignore
        return LabwareLike(parent)

    @property
    def is_well(self) -> bool:
        return self.object_type == LabwareLikeType.WELL

    @property
    def is_labware(self) -> bool:
        return self.object_type == LabwareLikeType.LABWARE

    @property
    def is_slot(self) -> bool:
        return self.object_type == LabwareLikeType.SLOT

    @property
    def is_empty(self) -> bool:
        return self.object_type == LabwareLikeType.NONE

    @property
    def is_module(self) -> bool:
        return self.object_type == LabwareLikeType.MODULE

    def as_well(self) -> "Well":
        # Import locally to avoid circular dependency
        from opentrons.protocol_api.labware import Well

        return cast(Well, self.object)

    def as_labware(self) -> "Labware":
        # Import locally to avoid circular dependency
        from opentrons.protocol_api.labware import Labware

        return cast(Labware, self.object)

    def as_module(self) -> "ModuleGeometry":
        from opentrons.protocols.geometry.module_geometry import ModuleGeometry

        return cast(ModuleGeometry, self.object)

    def get_parent_labware_and_well(
        self,
    ) -> Tuple[Optional["Labware"], Optional["Well"]]:
        """Attempt to split into a labware and well."""
        if self.is_labware:
            return self.as_labware(), None
        elif self.is_well and self.parent:
            return (
                self.parent.as_labware(),
                self.as_well(),
            )
        else:
            return None, None

    def first_parent(self) -> Optional[str]:
        """Return the topmost parent of this location. It should be
        either a string naming a slot or a None if the location isn't
        associated with a slot"""

        # cycle-detecting recursive climbing
        seen: Set[LabwareLike] = set()

        # internal function to have the cycle detector different per call
        def _fp_recurse(location: LabwareLike):
            if location in seen:
                raise RuntimeError("Cycle in labware parent")
            seen.add(location)
            if location.is_empty:
                return None
            elif location.is_slot:
                return str(location)
            else:
                return location.parent.first_parent()

        return _fp_recurse(self)

    def module_parent(self) -> Optional["ModuleGeometry"]:
        """
        Return the closest parent of this LabwareLike (including, possibly,
        the wrapped object) that is a ModuleGeometry
        """

        def recursive_get_module_parent(obj: LabwareLike) -> Optional["ModuleGeometry"]:
            if obj.is_module:
                return obj.as_module()
            next_obj = obj.parent
            if next_obj.is_empty:
                return None
            else:
                return recursive_get_module_parent(obj.parent)

        if self.is_empty:
            return None
        return recursive_get_module_parent(self)

    def quirks_from_any_parent(self) -> Set[str]:
        """Walk the tree of wells and labwares and extract quirks"""

        def recursive_get_quirks(obj: LabwareLike, found: List[str]) -> List[str]:
            if obj.is_labware:
                return found + obj.as_labware().quirks
            elif obj.is_well:
                return recursive_get_quirks(obj.parent, found)
            else:
                return found

        return set(recursive_get_quirks(self, []))

    def is_fixed_trash(self) -> bool:
        """Check if fixedTrash quirk is in any parent."""
        return "fixedTrash" in self.quirks_from_any_parent()

    def center_multichannel_on_wells(self) -> bool:
        """Check if centerMultichannelOnWells quirk is in any parent."""
        return "centerMultichannelOnWells" in self.quirks_from_any_parent()

    def __str__(self) -> str:
        return self._as_str

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, LabwareLike)
            and self.object == other.object
        )

    def __hash__(self):
        return id(self)



"""Graph Builder.

Generic graph builder classes.
"""
from __future__ import annotations

from typing import List, Dict, Callable, Sequence, Generic

from .types import VertexLike, VertexName


class Vertex(Generic[VertexName, VertexLike]):
    """Vertex class.

    A class to hold information about each vertex
    of a graph.
    """

    def __init__(self, vertex: VertexLike, neighbors: List[VertexName]) -> None:
        """Vertex class initializer.

        :param vertex: A node dataclass
        :param neighbors: A list of node names who
        are neighbors to the node dataclass
        """
        self._vertex = vertex
        self._neighbors = neighbors

    @property
    def name(self) -> VertexName:
        """Vertex class property: name.

        :returns: Name of the vertex
        """
        return self._vertex.name

    @property
    def vertex(self) -> VertexLike:
        """Vertex class property: vertex.

        :returns: The node dataclass
        """
        return self._vertex

    @property
    def neighbors(self) -> List[VertexName]:
        """Vertex class property: neighbors.

        :returns: The list of node names who are
        neighbors with the vertex.
        """
        return self._neighbors

    def add_neighbor(self, vertex_name: VertexName) -> None:
        """Add a neighbor.

        :param vertex_name: The name of the neighbor
        """
        self._neighbors.append(vertex_name)
        self._neighbors.sort()

    def remove_neighbor(self, vertex_name: VertexName) -> None:
        """Remove a neighbor.

        :param vertex_name: The name of the neighbor
        """
        if vertex_name in self._neighbors:
            self._neighbors.remove(vertex_name)


def default_sort(vertex: Vertex) -> VertexName:
    """Sort function default for a graph.

    By default, a graph's nodes will be searched
    by the name of the node. Generally, the name
    should either be a string or an integer.
    """
    return vertex.name


class Graph(Generic[VertexName, VertexLike]):
    """Graph class.

    A class to handle functions moving through the
    graph.
    """

    # Note, the type of sort_by is actually
    # Callable[[Vertex], VertexName] however there
    # is an issue when using generics for functions
    # passed into sort. See
    # https://github.com/python/typing/issues/760
    def __init__(
        self,
        sorted_graph: List[Vertex],
        lookup_table: Dict[VertexName, Vertex],
        sort_by: Callable[[Vertex], str],
    ) -> None:
        """Graph class initializer.

        :param sorted_graph: The initial graph, sorted
        and converted to vertex objects.
        :param lookup_table: A lookup table keyed by vertex
        name and with a value of the vertex object
        :param sort_by: The callable function used to sort
        the graph nodes in priority order.
        """
        self._sort_by = sort_by
        self._sorted_graph = sorted_graph
        self._lookup_table = lookup_table

    @classmethod
    def build(cls, graph: List[VertexLike], sort_by: Callable = default_sort) -> Graph:
        """Graph class builder.

        :param graph: A list of nodes to add to the graph.
        :param sort_by: The function used to sort the graph
        in priority order.
        :returns: A graph class
        """
        sorted_graph = []
        lookup_table = {}
        for vertex in graph:
            vertex_obj = cls.build_vertex(vertex)
            lookup_table[vertex.name] = vertex_obj
            sorted_graph.append(vertex_obj)
        sorted_graph.sort(key=sort_by)
        return cls(sorted_graph, lookup_table, sort_by)

    @property
    def graph(self) -> Sequence[Vertex]:
        """Graph class property: graph.

        :returns: A list of sorted vertex objects
        """
        return self._sorted_graph

    @staticmethod
    def build_vertex(vertex: VertexLike) -> Vertex:
        """Build a vertex.

        Use this to sort the neighbors and then build
        a vertex using a node dataclass.
        :param vertex: A node dataclass
        :returns: vertex object
        """
        vertex.sub_names.sort()
        return Vertex(vertex, vertex.sub_names)

    def add_vertex(self, vertex: VertexLike) -> None:
        """Add a vertex.

        :param vertex: A node dataclass
        """
        new_vertex = self.build_vertex(vertex)
        if new_vertex not in self._lookup_table.values():
            self._lookup_table[new_vertex.name] = new_vertex
            self._sorted_graph.append(new_vertex)
        self._sorted_graph.sort(key=self._sort_by)

    def remove_vertex(self, vertex: VertexLike) -> None:
        """Remove a vertex.

        :param vertex: A node dataclass
        """
        if vertex.name in self._lookup_table.keys():
            vertex_to_remove = self._lookup_table[vertex.name]
            del self._lookup_table[vertex.name]
            self._sorted_graph.remove(vertex_to_remove)

    def get_vertex(self, vertex_name: VertexName) -> Vertex:
        """Get a vertex.

        :param vertex_name: The name of the vertex
        :returns: The vertex object
        """
        return self._lookup_table[vertex_name]



"""Python Protocol API v3 type definitions and value classes."""
from opentrons_shared_data.labware.dev_types import LabwareParameters

from opentrons.types import (
    DeckSlotName,
    Location,
    MountType as Mount,
    Mount as DeprecatedMount,
    Point,
)

from opentrons.protocol_engine import DeckSlotLocation, PipetteName

__all__ = [
    # re-exports from opentrons_shared_data.labware.dev_types
    "LabwareParameters",
    # re-exports from opentrons.types
    "DeckSlotName",
    "Location",
    "Mount",
    "DeprecatedMount",
    "Point",
    # re-exports from opentrons.protocol_engine
    "DeckSlotLocation",
    "PipetteName",
]



"""Deprecated pipette module for the Python Protocol API.

Previous versions of the Python Protocol API referred to pipettes by
the name "instrument." This module exists to preserve backwards
compatibilty with old import paths.

This module will be removed in Protocol API v4.
"""
from .pipette_context import PipetteContext

InstrumentContext = PipetteContext
"""Alias of PipetteContext to preserve compatibility with Protocol API v2.

.. deprecated:: Protocol API v3.0
    Use :py:class:`PipetteContext` instead.
"""



from copy import deepcopy
import json
import logging
import os
from dataclasses import asdict
from pathlib import Path

from typing import Any, Dict, List, Union, Optional, TypeVar, cast

from opentrons.config import CONFIG
from opentrons.hardware_control.types import BoardRevision
from .types import CurrentDict, RobotConfig, AxisDict

log = logging.getLogger(__name__)

ROBOT_CONFIG_VERSION = 4

PLUNGER_CURRENT_LOW = 0.05
PLUNGER_CURRENT_HIGH = 0.05

MOUNT_CURRENT_LOW = 0.1
MOUNT_CURRENT_HIGH = 0.8

X_CURRENT_LOW = 0.3
X_CURRENT_HIGH = 1.25

Y_CURRENT_LOW = 0.3
Y_CURRENT_HIGH = 1.25

XY_CURRENT_LOW_REFRESH = 0.7
MOUNT_CURRENT_HIGH_REFRESH = 0.5

Z_RETRACT_DISTANCE = 2

HIGH_CURRENT: CurrentDict = {
    "default": {
        "X": X_CURRENT_HIGH,
        "Y": Y_CURRENT_HIGH,
        "Z": MOUNT_CURRENT_HIGH_REFRESH,
        "A": MOUNT_CURRENT_HIGH_REFRESH,
        "B": PLUNGER_CURRENT_HIGH,
        "C": PLUNGER_CURRENT_HIGH,
    },
    "2.1": {
        "X": X_CURRENT_HIGH,
        "Y": Y_CURRENT_HIGH,
        "Z": MOUNT_CURRENT_HIGH,
        "A": MOUNT_CURRENT_HIGH,
        "B": PLUNGER_CURRENT_HIGH,
        "C": PLUNGER_CURRENT_HIGH,
    },
}

LOW_CURRENT: CurrentDict = {
    "default": {
        "X": XY_CURRENT_LOW_REFRESH,
        "Y": XY_CURRENT_LOW_REFRESH,
        "Z": MOUNT_CURRENT_LOW,
        "A": MOUNT_CURRENT_LOW,
        "B": PLUNGER_CURRENT_LOW,
        "C": PLUNGER_CURRENT_LOW,
    },
    "2.1": {
        "X": X_CURRENT_LOW,
        "Y": Y_CURRENT_LOW,
        "Z": MOUNT_CURRENT_LOW,
        "A": MOUNT_CURRENT_LOW,
        "B": PLUNGER_CURRENT_LOW,
        "C": PLUNGER_CURRENT_LOW,
    },
}

DEFAULT_CURRENT: CurrentDict = {
    "default": {
        "X": HIGH_CURRENT["default"]["X"],
        "Y": HIGH_CURRENT["default"]["Y"],
        "Z": HIGH_CURRENT["default"]["Z"],
        "A": HIGH_CURRENT["default"]["A"],
        "B": LOW_CURRENT["default"]["B"],
        "C": LOW_CURRENT["default"]["C"],
    },
    "2.1": {
        "X": HIGH_CURRENT["2.1"]["X"],
        "Y": HIGH_CURRENT["2.1"]["Y"],
        "Z": HIGH_CURRENT["2.1"]["Z"],
        "A": HIGH_CURRENT["2.1"]["A"],
        "B": LOW_CURRENT["2.1"]["B"],
        "C": LOW_CURRENT["2.1"]["C"],
    },
}

X_MAX_SPEED = 600
Y_MAX_SPEED = 400
Z_MAX_SPEED = 125
A_MAX_SPEED = 125
B_MAX_SPEED = 40
C_MAX_SPEED = 40

DEFAULT_MAX_SPEEDS: AxisDict = {
    "X": X_MAX_SPEED,
    "Y": Y_MAX_SPEED,
    "Z": Z_MAX_SPEED,
    "A": A_MAX_SPEED,
    "B": B_MAX_SPEED,
    "C": C_MAX_SPEED,
}

DEFAULT_CURRENT_STRING = " ".join(
    ["{}{}".format(key, value) for key, value in DEFAULT_CURRENT.items()]
)

DEFAULT_DECK_CALIBRATION_V2: List[List[float]] = [
    [1.00, 0.00, 0.00],
    [0.00, 1.00, 0.00],
    [0.00, 0.00, 1.00],
]

DEFAULT_SIMULATION_CALIBRATION: List[List[float]] = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, -25.0],
    [0.0, 0.0, 0.0, 1.0],
]

X_ACCELERATION = 3000
Y_ACCELERATION = 2000
Z_ACCELERATION = 1500
A_ACCELERATION = 1500
B_ACCELERATION = 200
C_ACCELERATION = 200

DEFAULT_ACCELERATION: Dict[str, float] = {
    "X": X_ACCELERATION,
    "Y": Y_ACCELERATION,
    "Z": Z_ACCELERATION,
    "A": A_ACCELERATION,
    "B": B_ACCELERATION,
    "C": C_ACCELERATION,
}

DEFAULT_PIPETTE_CONFIGS: Dict[str, float] = {
    "homePosition": 220,
    "stepsPerMM": 768,
    "maxTravel": 30,
}

DEFAULT_GANTRY_STEPS_PER_MM: Dict[str, float] = {
    "X": 80.00,
    "Y": 80.00,
    "Z": 400,
    "A": 400,
}


DEFAULT_MOUNT_OFFSET = [-34, 0, 0]
DEFAULT_PIPETTE_OFFSET = [0.0, 0.0, 0.0]
SERIAL_SPEED = 115200
DEFAULT_LOG_LEVEL = "INFO"


def _build_hw_versioned_current_dict(
    from_conf: Optional[Dict[str, Any]], default: CurrentDict
) -> CurrentDict:
    if not from_conf or not isinstance(from_conf, dict):
        return default
    # special case: if this is a valid old (i.e. not model-specific) current
    # setup, migrate it.
    if "default" not in from_conf and not (set("XYZABC") - set(from_conf.keys())):
        new_dct = deepcopy(default)
        # Because there's no case in which a machine with a more recent revision
        # than 2.1 should have a valid and edited robot config when updating
        # to this code, we should default it to 2.1 to avoid breaking other
        # robots
        new_dct["2.1"] = cast(AxisDict, from_conf)
        return new_dct
    return cast(CurrentDict, from_conf)


DictType = TypeVar("DictType", bound=Dict)


def _build_dict_with_default(
    from_conf: Union[DictType, str, None], default: DictType
) -> DictType:
    if not isinstance(from_conf, dict):
        return default
    else:
        return cast(DictType, from_conf)


def current_for_revision(
    current_dict: CurrentDict, revision: BoardRevision
) -> AxisDict:
    if revision == BoardRevision.UNKNOWN:
        return current_dict.get("2.1", current_dict["default"])
    elif revision.real_name() in current_dict:
        return current_dict[revision.real_name()]  # type: ignore
    else:
        return current_dict["default"]


def build_config(robot_settings: Dict[str, Any]) -> RobotConfig:
    return RobotConfig(
        name=robot_settings.get("name", "Ada Lovelace"),
        version=ROBOT_CONFIG_VERSION,
        gantry_steps_per_mm=_build_dict_with_default(
            robot_settings.get("steps_per_mm"), DEFAULT_GANTRY_STEPS_PER_MM
        ),
        acceleration=_build_dict_with_default(
            robot_settings.get("acceleration"), DEFAULT_ACCELERATION
        ),
        serial_speed=robot_settings.get("serial_speed", SERIAL_SPEED),
        default_current=_build_hw_versioned_current_dict(
            robot_settings.get("default_current"), DEFAULT_CURRENT
        ),
        low_current=_build_hw_versioned_current_dict(
            robot_settings.get("low_current"), LOW_CURRENT
        ),
        high_current=_build_hw_versioned_current_dict(
            robot_settings.get("high_current"), HIGH_CURRENT
        ),
        default_max_speed=robot_settings.get("default_max_speed", DEFAULT_MAX_SPEEDS),
        log_level=robot_settings.get("log_level", DEFAULT_LOG_LEVEL),
        default_pipette_configs=robot_settings.get(
            "default_pipette_configs", DEFAULT_PIPETTE_CONFIGS
        ),
        z_retract_distance=robot_settings.get("z_retract_distance", Z_RETRACT_DISTANCE),
        left_mount_offset=robot_settings.get("left_mount_offset", DEFAULT_MOUNT_OFFSET),
    )


def config_to_save(config: RobotConfig) -> Dict[str, Any]:
    return asdict(config)


def load() -> RobotConfig:
    settings_file = CONFIG["robot_settings_file"]
    log.debug("Loading robot settings from {}".format(settings_file))
    robot_settings = _load_json(settings_file) or {}
    return build_config(robot_settings)


def save_robot_settings(config: RobotConfig, rs_filename: str = None, tag: str = None):
    config_dict = config_to_save(config)

    # Save everything else in a different file
    filename = rs_filename or CONFIG["robot_settings_file"]
    if tag:
        root, ext = os.path.splitext(filename)
        filename = "{}-{}{}".format(root, tag, ext)
    _save_json(config_dict, filename=filename)

    return config_dict


def backup_configuration(config: RobotConfig, tag: str = None) -> None:
    import time

    if not tag:
        tag = str(int(time.time() * 1000))
    save_robot_settings(config, tag=tag)


def get_legacy_gantry_calibration() -> Optional[List[List[float]]]:
    """
    Returns the legacy gantry calibration if exists.

    This should happen only if the new deck calibration file does not exist.
    The legacy calibration should then be migrated to the new format.
    """
    gantry_cal = _load_json(CONFIG["deck_calibration_file"])
    if "gantry_calibration" in gantry_cal:
        return gantry_cal["gantry_calibration"]
    else:
        return None


def clear() -> None:
    _clear_file(CONFIG["robot_settings_file"])


def _clear_file(filename: Union[str, Path]) -> None:
    log.debug("Deleting {}".format(filename))
    if os.path.exists(filename):
        os.remove(filename)


# TODO: move to util (write a default load, save JSON function)
def _load_json(filename: Union[str, Path]) -> Dict[str, Any]:
    try:
        with open(filename, "r") as file:
            res = json.load(file)
    except FileNotFoundError:
        log.warning("{0} not found. Loading defaults".format(filename))
        res = {}
    except json.decoder.JSONDecodeError:
        log.warning("{0} is corrupt. Loading defaults".format(filename))
        res = {}
    return res


def _save_json(data: Dict[str, Any], filename: Union[str, Path]) -> None:
    try:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as file:
            json.dump(data, file, sort_keys=True, indent=4)
            file.flush()
            os.fsync(file.fileno())
    except OSError:
        log.exception("Write failed with exception:")



import asyncio
from opentrons.drivers.smoothie_drivers import SmoothieDriver

from . import args_handler

BAD_BARCODE_MESSAGE = "Unexpected Serial -> {}"
WRITE_FAIL_MESSAGE = "Data not saved, HOLD BUTTON"

# must iterate through v1.4 and v1.3 first, because v1 barcodes did not
# have characters to specify the version number
VERSIONS = ["v2.2", "v2.1", "v2", "v1.6", "v1.5", "v1.4", "v1.3", "v1"]

MODELS = {
    "v1": {
        "P10S": "p10_single_v1",
        "P10M": "p10_multi_v1",
        "P50S": "p50_single_v1",
        "P50M": "p50_multi_v1",
        "P300S": "p300_single_v1",
        "P300M": "p300_multi_v1",
        "P1000S": "p1000_single_v1",
    },
    "v1.3": {
        "P10SV13": "p10_single_v1.3",
        "P10MV13": "p10_multi_v1.3",
        "P50SV13": "p50_single_v1.3",
        "P50MV13": "p50_multi_v1.3",
        "P3HSV13": "p300_single_v1.3",
        "P3HMV13": "p300_multi_v1.3",
        "P1KSV13": "p1000_single_v1.3",
    },
    "v1.4": {
        "P10SV14": "p10_single_v1.4",
        "P10MV14": "p10_multi_v1.4",
        "P50SV14": "p50_single_v1.4",
        "P50MV14": "p50_multi_v1.4",
        "P3HSV14": "p300_single_v1.4",
        "P3HMV14": "p300_multi_v1.4",
        "P1KSV14": "p1000_single_v1.4",
    },
    "v1.5": {
        "P10SV15": "p10_single_v1.5",
        "P10MV15": "p10_multi_v1.5",
        "P50SV15": "p50_single_v1.5",
        "P50MV15": "p50_multi_v1.5",
        "P3HSV15": "p300_single_v1.5",
        "P3HMV15": "p300_multi_v1.5",
        "P1KSV15": "p1000_single_v1.5",
    },
    "v1.6": {
        "P10MV16": "p10_multi_v1.6",
    },
    "v2": {
        "P3HSV20": "p300_single_v2.0",
        "P3HMV20": "p300_multi_v2.0",
        "P1KSV20": "p1000_single_v2.0",
        "P20SV20": "p20_single_v2.0",
        "P20MV20": "p20_multi_v2.0",
    },
    "v2.1": {
        "P3HSV21": "p300_single_v2.1",
        "P1KSV21": "p1000_single_v2.1",
        "P20SV21": "p20_single_v2.1",
        "P3HMV21": "p300_multi_v2.1",
        "P20MV21": "p20_multi_v2.1",
    },
    "v2.2": {
        "P20MV22": "p20_multi_v2.2",
        "P20SV22": "p20_single_v2.2",
        "P1KSV22": "p1000_single_v2.2",
    },
}


async def write_identifiers(
    mount: str, new_id: str, new_model: str, driver: SmoothieDriver
) -> None:
    """
    Send a bytearray to the specified mount, so that Smoothieware can
    save the bytes to the pipette's memory
    """
    await driver.write_pipette_id(mount, new_id)
    read_id = await driver.read_pipette_id(mount)
    _assert_the_same(new_id, read_id)
    await driver.write_pipette_model(mount, new_model)
    read_model = await driver.read_pipette_model(mount)
    _assert_the_same(new_model, read_model)


async def check_previous_data(mount: str, driver: SmoothieDriver) -> None:
    old_id = await driver.read_pipette_id(mount)
    old_model = await driver.read_pipette_model(mount)
    if old_id and old_model:
        print("Overwriting old data: id={0}, model={1}".format(old_id, old_model))
    else:
        print("No old data on this pipette")


def _assert_the_same(a, b):
    if a != b:
        raise Exception(WRITE_FAIL_MESSAGE)


def _user_submitted_barcode(max_length: int) -> str:
    """
    User can enter a serial number as a string of HEX values
    Length of byte array must equal `num`
    """
    barcode = input("BUTTON + SCAN: ").strip()
    if len(barcode) > max_length:
        raise Exception(BAD_BARCODE_MESSAGE.format(barcode))
    # remove all characters before the letter P
    # for example, remove ASCII selector code "\x1b(B" on chinese keyboards
    barcode = barcode[barcode.index("P") :]
    barcode = barcode.split("\r")[0].split("\n")[0]  # remove any newlines
    return barcode


def _parse_model_from_barcode(barcode: str) -> str:
    for version in VERSIONS:
        for barcode_substring in MODELS[version].keys():
            if barcode.startswith(barcode_substring):
                return MODELS[version][barcode_substring]
    raise Exception(BAD_BARCODE_MESSAGE.format(barcode))


async def _do_wpm(mount: str, driver: SmoothieDriver) -> None:
    try:
        barcode = _user_submitted_barcode(32)
        model = _parse_model_from_barcode(barcode)
        await check_previous_data(mount, driver)
        await write_identifiers(mount, barcode, model, driver)
        print("PASS: Saved -> {0} (model {1})".format(barcode, model))
    except KeyboardInterrupt:
        exit()
    except Exception as e:
        print("FAIL: {}".format(e))


async def main() -> None:
    parser = args_handler.root_argparser("Write model and serial to a pipette's eeprom")
    parser.add_argument(
        "-m",
        "--mount",
        help="mount to use",
        default="right",
        type=str,
        choices=["left", "right"],
    )
    args = parser.parse_args()
    _, driver = await args_handler.build_driver(args.port)
    while True:
        await _do_wpm(args.mount, driver)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Dict
from opentrons.protocol_api.labware import Labware
from opentrons.protocol_api.contexts import InstrumentContext


Instruments = Dict[str, InstrumentContext]

LoadedLabware = Dict[str, Labware]



import logging
from typing import Dict

from opentrons.drivers.smoothie_drivers.constants import AXES, GCODE_ROUNDING_PRECISION
from opentrons.drivers.utils import (
    parse_key_values,
    parse_number,
    parse_optional_number,
    ParseError,
)


log = logging.getLogger(__name__)


def parse_position_response(raw_axis_values: str) -> Dict[str, float]:
    """
    Parse position response.

    Args:
        raw_axis_values: a string containing axis:value

    Returns:
        dictionary of axis to position.
    """
    parsed_values = parse_key_values(raw_axis_values)
    if len(parsed_values) < 6:
        raise ParseError(
            error_message="Unexpected response in _parse_position_response",
            parse_source=raw_axis_values,
        )

    data = {
        k.title(): parse_number(v, GCODE_ROUNDING_PRECISION)
        for k, v in parsed_values.items()
    }
    return data


def parse_instrument_data(smoothie_response: str) -> Dict[str, bytearray]:
    """
    Parse instrument data.

    Args:
        smoothie_response: A string containing a mount prefix (L or R) followed by :
            and a hex string.

    Returns:
        mapping of the mount prefix to the hex string.
    """
    try:
        items = smoothie_response.split("\n")[0].strip().split(":")
        mount = items[0]
        if mount not in {"L", "R"}:
            raise ParseError(
                error_message=f"Invalid mount '{mount}'", parse_source=smoothie_response
            )
        # data received from Smoothieware is stringified HEX values
        # because of how Smoothieware handles GCODE messages
        data = bytearray.fromhex(items[1])
    except (ValueError, IndexError, TypeError, AttributeError):
        raise ParseError(
            error_message="Unexpected argument to parse_instrument_data",
            parse_source=smoothie_response,
        )
    return {mount: data}


def byte_array_to_ascii_string(byte_array: bytearray) -> str:
    """
    Convert byte array to ascii string.

    Args:
        byte_array: a byte array

    Returns:
        String

    """
    # remove trailing null characters
    try:
        for c in [b"\x00", b"\xFF"]:
            if c in byte_array:
                byte_array = byte_array[: byte_array.index(c)]
        res = byte_array.decode()
    except (ValueError, TypeError, AttributeError):
        log.exception("Unexpected argument to _byte_array_to_ascii_string:")
        raise ParseError(
            error_message="Unexpected argument to byte_array_to_ascii_string",
            parse_source=byte_array.decode(),
        )
    return res


def parse_switch_values(raw_switch_values: str) -> Dict[str, bool]:
    if not raw_switch_values or not isinstance(raw_switch_values, str):
        raise ParseError(
            error_message="Unexpected argument to parse_switch_values",
            parse_source=raw_switch_values,
        )

    # probe has a space after it's ":" for some reason
    if "Probe: " in raw_switch_values:
        raw_switch_values = raw_switch_values.replace("Probe: ", "Probe:")

    parsed_values = parse_key_values(raw_switch_values)
    res = {
        k.title(): bool(parse_optional_number(v, rounding_val=GCODE_ROUNDING_PRECISION))
        for (k, v) in parsed_values.items()
        if any(n in k for n in ["max", "Probe"])
    }
    # remove the extra "_max" character from each axis key in the dict
    res = {key.split("_")[0]: val for key, val in res.items()}
    if len((list(AXES) + ["Probe"]) & res.keys()) != 7:
        raise ParseError(
            error_message="Unexpected argument to parse_switch_values",
            parse_source=raw_switch_values,
        )
    return res


def parse_homing_status_values(raw_homing_status_values):
    """
    Parse the Smoothieware response to a G28.6 command (homing-status)
    A "1" means it has been homed, and "0" means it has not been homed

    Example response after homing just X axis:
    "X:1 Y:0 Z:0 A:0 B:0 C:0"

    returns: dict
        Key is axis, value is True if the axis needs to be homed
    """
    parsed_values = parse_key_values(raw_homing_status_values)
    res = {
        k.title(): bool(parse_number(v, GCODE_ROUNDING_PRECISION))
        for k, v in parsed_values.items()
    }
    if len(list(AXES) & res.keys()) != 6:
        raise ParseError(
            error_message="Unexpected argument to parse_homing_status_values",
            parse_source=raw_homing_status_values,
        )
    return res



"""Protocol run control and management.

The main export of this module is the ProtocolRunner class. See
protocol_runner.py for more details.
"""
from .protocol_runner import ProtocolRunner, ProtocolRunData
from .protocol_file import ProtocolFile, ProtocolFileType
from .create_simulating_runner import create_simulating_runner

__all__ = [
    "ProtocolRunner",
    "ProtocolRunData",
    "ProtocolFile",
    "ProtocolFileType",
    "create_simulating_runner",
]



import functools
from abc import ABC, abstractmethod
from typing import Any, cast, Callable, TypeVar

from opentrons.config.feature_flags import enable_http_protocol_sessions
from opentrons.hardware_control import ThreadedAsyncLock


class RobotBusy(ABC):
    @property
    @abstractmethod
    def busy_lock(self) -> ThreadedAsyncLock:
        ...


Func = TypeVar("Func", bound=Callable[..., Any])


def robot_is_busy(func: Func) -> Func:
    """Decorator to mark a function as putting the robot in a busy
    state. Simultaneous attempts to call a busy function (from same or
    separate thread) will result in an exception.
     Must wrap a class of type RobotBusy.

     :raises ThreadedAsyncForbidden: on call during busy.
    """

    @functools.wraps(func)
    def decorated(*args: Any, **kwargs: Any) -> Any:
        self = args[0]
        if self.busy_lock:
            with self.busy_lock.forbid():
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return cast(Func, decorated)


def requires_http_protocols_disabled(func: Func) -> Func:
    """Decorator that makes sure the enableHttpProtocolSessions is disabled
    before proceeding.

    :raises RuntimeError: if enableHttpProtocolSessions is enabled
    """

    @functools.wraps(func)
    def decorated(*args: Any, **kwargs: Any) -> Any:
        if enable_http_protocol_sessions():
            raise RuntimeError(
                "Please disable the 'Enable Experimental HTTP Protocol "
                "Sessions' advanced setting for this robot if you'd like to "
                "upload protocols from the Opentrons App"
            )
        return func(*args, **kwargs)

    return cast(Func, decorated)



import typing

from opentrons.protocols.advanced_control.transfers import MixStrategy, Mix


def mix_from_kwargs(
    top_kwargs: typing.Dict[str, typing.Any]
) -> typing.Tuple[MixStrategy, Mix]:
    """A utility function to determine mix strategy from key word arguments
    to InstrumentContext.mix"""

    def _mix_requested(kwargs, opt):
        """
        Helper for determining mix options from :py:meth:`transfer` kwargs
        Mixes can be ignored in kwargs by either
        - Not specifying the kwarg
        - Specifying it as None
        - Specifying it as (0, 0)

        This handles all these cases.
        """
        val = kwargs.get(opt)
        if None is val:
            return False
        if val == (0, 0):
            return False
        return True

    mix_opts = Mix()
    if _mix_requested(top_kwargs, "mix_before") and _mix_requested(
        top_kwargs, "mix_after"
    ):
        mix_strategy = MixStrategy.BOTH
        before_opts = top_kwargs["mix_before"]
        after_opts = top_kwargs["mix_after"]
        mix_opts = mix_opts._replace(
            mix_after=mix_opts.mix_after._replace(
                repetitions=after_opts[0], volume=after_opts[1]
            ),
            mix_before=mix_opts.mix_before._replace(
                repetitions=before_opts[0], volume=before_opts[1]
            ),
        )
    elif _mix_requested(top_kwargs, "mix_before"):
        mix_strategy = MixStrategy.BEFORE
        before_opts = top_kwargs["mix_before"]
        mix_opts = mix_opts._replace(
            mix_before=mix_opts.mix_before._replace(
                repetitions=before_opts[0], volume=before_opts[1]
            )
        )
    elif _mix_requested(top_kwargs, "mix_after"):
        mix_strategy = MixStrategy.AFTER
        after_opts = top_kwargs["mix_after"]
        mix_opts = mix_opts._replace(
            mix_after=mix_opts.mix_after._replace(
                repetitions=after_opts[0], volume=after_opts[1]
            )
        )
    else:
        mix_strategy = MixStrategy.NEVER
    return mix_strategy, mix_opts



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from opentrons import types
from opentrons.protocol_api.labware import Labware, Well


class AbstractPairedInstrument(ABC):
    @abstractmethod
    def pick_up_tip(
        self,
        target: Well,
        secondary_target: Well,
        tiprack: Labware,
        presses: Optional[int],
        increment: Optional[float],
        tip_length: float,
    ) -> None:
        ...

    @abstractmethod
    def drop_tip(self, target: types.Location, home_after: bool) -> None:
        ...

    @abstractmethod
    def move_to(
        self,
        location: types.Location,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def aspirate(
        self,
        volume: float,
        location: types.Location,
        rate: float,
    ) -> None:
        ...

    @abstractmethod
    def dispense(self, volume: float, location: types.Location, rate: float) -> None:
        ...

    @abstractmethod
    def blow_out(self, location: types.Location) -> None:
        ...

    @abstractmethod
    def touch_tip(
        self, well: Well, radius: float, v_offset: float, speed: float
    ) -> None:
        ...
